from typing import Iterable
from enum import IntEnum
import numpy as np
from ..framework.Edit import Allele, Edit

AA_SET = {'A', 'C', 'D', 'E', 'F', 
'G', 'H', 'I', 'K', 'L', 
'M', 'N', 'P', 'Q', 'R',
 'S', 'T', 'V', 'W', 'Y', '*', '/'}

def jaccard(set1, set2):
    intersection = len(set(set1).intersection(set2))
    union = (len(set1) + len(set2)) - intersection
    if union == 0: return np.nan
    return float(intersection) / union

class MutationType(IntEnum):
    NO_CHANGE = -1
    SYNONYMOUS = 0
    MISSENSE = 1
    NONSENSE = 2

class AminoAcidEdit(Edit):
    def __init__(self, pos: int, ref: str, alt: str):
        self.pos = int(pos)
        assert ref in AA_SET, "Invalid ref aa: {}".format(ref)
        assert alt in AA_SET, "Invalid alt aa: {}".format(alt)
        self.ref = ref
        self.alt = alt

    @classmethod
    def from_str(cls, edit_str):
        pos, aa_change = edit_str.split(":")
        ref, alt = aa_change.split(">")
        return(cls(pos, ref, alt))

    def get_abs_edit(self):
        return("A"+self.__repr__())

    def __repr__(self):
        return("{}:{}>{}".format(int(self.pos),
        self.ref, self.alt))

    def __hash__(self):
        return(hash(self.__repr__()))

    def _severity(self):
        if self.ref == self.alt: return(MutationType.SYNONYMOUS)
        if self.alt == "*": return(MutationType.NONSENSE)
        if self.alt in AA_SET: return(MutationType.MISSENSE)
        else:
            raise ValueError("Alt base invalid:{}".format(self.alt))

    def __eq__(self, other):
        if (
            self.pos == other.pos
            and self.ref == other.ref
            and self.alt == other.alt
        ):
            return True
        return False

    def __lt__(self, other): # Implemented for pandas compatibility
        return self.pos < other.pos

    def __gt__(self, other): # Implemented for pandas compatibility
        return self.pos > other.pos

class AminoAcidAllele(Allele):
    def __init__(self, edits: Iterable[AminoAcidEdit] = None):
        if edits is None:
            self.edits = set()
        else:
            self.edits = set(edits)

    @classmethod
    def from_str(cls, allele_str): #pos:strand:start>end
        if type(allele_str) is AminoAcidAllele: return(allele_str)
        edits = set()
        try:
            for edit_str in allele_str.split(","):
                edit = AminoAcidEdit.from_str(edit_str)
                edits.add(edit)
        except ValueError:
            if allele_str.strip() == '':
                return(cls(None))
        return(cls(edits))

    def get_most_severe(self):
        if len(self.edits) == 0:
            return(MutationType.NO_CHANGE)
        severities = []
        for edit in self.edits:
            severities.append(edit._severity())
        return(max(severities))


class CodingNoncodingAllele():
    def __init__(self, aa_edits: Iterable[AminoAcidEdit] = None, base_edits: Iterable[Edit] = None, unique_identifier = None):
        self.aa_allele = AminoAcidAllele(aa_edits)
        self.nt_allele = Allele(base_edits)
        self.uid = unique_identifier
        if not self.uid is None:
            self.set_uid(self.uid)

    @classmethod
    def from_str(cls, allele_str):
        if isinstance(allele_str, cls): return(allele_str)
        uid = None
        if '!' in allele_str:
            uid, allele_str = allele_str.split('!')
        try:
            aa_allele_str, nt_allele_str = allele_str.split("|")
            aa_allele = AminoAcidAllele.from_str(aa_allele_str)
            nt_allele = Allele.from_str(nt_allele_str)
        except ValueError:
            if allele_str.strip() == '':
                return(cls(None))
            else:
                print(allele_str)
        return(cls(aa_allele.edits, nt_allele.edits, unique_identifier = uid))

    def get_most_severe(self):
        aa_sev = self.aa_allele.get_most_severe()
        if len(self.nt_allele.edits) > 0: return max(aa_sev, 0.1)
        return aa_sev

    def set_uid(self, uid):
        self.uid = uid
        self.nt_allele.edits = set([e.set_uid(uid) for e in self.nt_allele.edits])

    def get_jaccard(self, other):
        aa_jaccard = jaccard(self.aa_allele.edits, other.aa_allele.edits)
        nt_jaccard = jaccard(self.nt_allele.edits, other.nt_allele.edits)
        return(aa_jaccard, nt_jaccard)

    def __repr__(self):
        if not self.uid is None:
            return self.uid + "!" + str(self.aa_allele) + "|" + str(self.nt_allele)
        return str(self.aa_allele) + "|" + str(self.nt_allele)

    def __eq__(self, other):
        if not isinstance(other, type(self)): return False
        return(self.uid == other.uid and self.aa_allele == other.aa_allele and self.nt_allele == other.nt_allele)
    
    def __lt__(self, other):
        if self.aa_allele < other.aa_allele: return True
        if self.aa_allele == other.aa_allele:
            if self.nt_allele < other.nt_allele: return True
        return False

    def __gt__(self, other):
        if self.aa_allele > other.aa_allele: return True
        if self.aa_allele == other.aa_allele:
            if self.nt_allele > other.nt_allele: return True
        return False

    def __hash__(self):
        return(hash(self.__repr__()))





from typing import Set, Iterable
import os
from os import path
import sys
import logging
from tqdm import tqdm
import pdb
import gzip
import numpy as np
import pandas as pd
from Bio import SeqIO
from Bio.SeqIO.QualityIO import FastqPhredIterator
from beret import ReporterScreen, Edit, Allele

# from beret.framework.Edit import Allele

from ._supporting_fn import (
    _base_edit_to_from,
    _read_count_match,
    _get_fastq_handle,
    _write_paired_end_reads,
    _get_edited_allele_crispresso,
    _check_readname_match,
    _read_is_good_quality,
    revcomp,
    _multiindex_dict_to_df,
    _write_alignment_matrix
)

logging.basicConfig(
    level=logging.INFO,
    format="%(levelname)-5s @ %(asctime)s:\n\t %(message)s \n",
    datefmt="%a, %d %b %Y %H:%M:%S",
    stream=sys.stderr,
    filemode="w",
)
error = logging.critical
warn = logging.warning
debug = logging.debug
info = logging.info


class NTException(Exception):
    pass


class InputFileError(Exception):
    pass


class NoReadsAfterQualityFiltering(Exception):
    pass


class GuideEditCounter:
    def __init__(self, **kwargs):
        self.R1_filename = kwargs["R1"]
        self.R2_filename = kwargs["R2"]
        self.base_edited_from = kwargs["edited_base"]
        self.base_edited_to = _base_edit_to_from(self.base_edited_from)
        self.min_average_read_quality = kwargs["min_average_read_quality"]
        self.min_single_bp_quality = kwargs["min_single_bp_quality"]

        self.guides_info_df = pd.read_csv(kwargs["sgRNA_filename"])
        self.guides_has_strands = ("Strand" in self.guides_info_df.columns) or (
            "strand" in self.guides_info_df.columns
        )
        if self.guides_has_strands:
            info("Considering strand information of guides")
            assert "start_pos" in self.guides_info_df.columns
            # assert "gRNA_end" in self.guides_info_df.columns
        else:
            info("Ignoring guide strands, all guides are considered positive")

        self.qstart_R1 = kwargs["qstart_R1"]
        self.qend_R1 = kwargs["qend_R1"]
        self.qstart_R2 = kwargs["qstart_R2"]
        self.qend_R2 = kwargs["qend_R2"]
        self.name = kwargs["name"]

        self.sgRNA_filename = kwargs["sgRNA_filename"]
        self.count_only_bcmatched = False
        self._set_sgRNA_df()

        self.database_id = self._get_database_name()
        self.output_dir = os.path.join(
            os.path.abspath(kwargs["output_folder"]),
            "beret_count_%s" % self.database_id,
        )
        self._write_start_log()

        self.screen = ReporterScreen(
            X=np.zeros((len(self.guides_info_df), 1)),
            X_edit=np.zeros((len(self.guides_info_df), 1)),
            X_bcmatch=np.zeros((len(self.guides_info_df), 1)),
            guides=self.guides_info_df,
            condit=pd.DataFrame(index=[self.database_id]),
        )
        self.count_guide_edits = kwargs["count_guide_edits"]
        if self.count_guide_edits:
            self.screen.uns["guide_edit_counts"] = dict()
        self.count_reporter_edits = kwargs["count_reporter"]
        if self.count_reporter_edits:
            self.screen.uns["edit_counts"] = dict()
            self.gstart_reporter = kwargs["gstart_reporter"]

        self.count_edited_alleles = kwargs["count_allele"]
        if self.count_edited_alleles:
            self.screen.uns["allele_counts"] = dict()
            # pd.DataFrame(
            #     columns = ["guide_name", "allele_id", "count"]).set_index(
            #     ["guide_name", "allele_id"], drop = True)
            #self.screen.uns["alleles"] = pd.DataFrame(columns=["allele_id", "edit_id"])
            # Dict[guide_name -> List[edit_ID]]

        self.count_guide_reporter_alleles = kwargs["count_guide_reporter_alleles"]
        if self.count_guide_reporter_alleles:
            self.guide_to_guide_reporter_allele = dict()

        self.guide_start_seq = kwargs["guide_start_seq"]
        self.guide_bc = kwargs["guide_bc"]
        if self.guide_bc:
            self.guide_bc_len = kwargs["guide_bc_len"]

        self.offset = kwargs["offset"]
        if self.count_reporter_edits or self.count_edited_alleles or self.count_guide_reporter_alleles:
            self.reporter_length = kwargs["reporter_length"]
        self.guide_to_allele = dict()
        self.n_total_reads = _read_count_match(self.R1_filename, self.R2_filename)

        self.keep_intermediate = kwargs["keep_intermediate"]
        self.semimatch = 0
        self.bcmatch = 0
        self.nomatch = 0
        self.duplicate_match = 0
        self.duplicate_match_wo_barcode = 0

    def masked_equal(self, seq1, seq2):
        return(seq1.replace(self.base_edited_from, self.base_edited_to) == \
            seq2.replace(self.base_edited_from, self.base_edited_to))

    def _set_sgRNA_df(self):
        with open(self.sgRNA_filename) as infile:
            sgRNA_df = pd.read_csv(infile)
            if not ("name" in sgRNA_df.columns and "sequence" in sgRNA_df.columns):
                raise InputFileError(
                    "Input gRNA info file doesn't have the column 'name' or 'sequence'."
                )
            if self.count_only_bcmatched and "barcode" not in sgRNA_df.columns:
                raise InputFileError(
                    "Input gRNA info file doesn't have the column 'barcode'."
                )
            sgRNA_df = sgRNA_df.set_index("name")
        self.guides_info_df = sgRNA_df
        self.guides_info_df['masked_sequence'] = self.guides_info_df.sequence.map(lambda s: s.replace(self.base_edited_from, self.base_edited_to))
        self.guides_info_df['masked_barcode'] = self.guides_info_df.barcode.map(lambda s: s.replace(self.base_edited_from, self.base_edited_to))
        self.guide_lengths = sgRNA_df.sequence.map(lambda s: len(s)).unique()


    def check_filter_fastq(self):
        self.filtered_R1_filename = self._jp(
            os.path.basename(self.R1_filename).replace(".fastq", "").replace(".gz", "")
            + "_filtered.fastq.gz"
        )
        self.filtered_R2_filename = self._jp(
            os.path.basename(self.R2_filename).replace(".fastq", "").replace(".gz", "")
            + "_filtered.fastq.gz"
        )
        if path.exists(self.filtered_R1_filename) and path.exists(
            self.filtered_R2_filename
        ):
            warn("Using preexisting filtered file")
        else:
            self._check_names_filter_fastq()

    def get_counts(self):
        infile_R1 = _get_fastq_handle(self.R1_filename)
        infile_R2 = _get_fastq_handle(self.R2_filename)

        self.nomatch_R1_filename = self.R1_filename.replace(".fastq", "_nomatch.fastq")
        self.nomatch_R2_filename = self.R2_filename.replace(".fastq", "_nomatch.fastq")
        self.semimatch_R1_filename = self.R1_filename.replace(
            ".fastq", "_semimatch.fastq"
        )
        self.semimatch_R2_filename = self.R2_filename.replace(
            ".fastq", "_semimatch.fastq"
        )
        if self.count_edited_alleles or self.count_guide_reporter_alleles:
            _write_alignment_matrix(self.base_edited_from, self.base_edited_to, self.output_dir + "/.aln_mat.txt")

        if self.count_only_bcmatched:
            # count X
            self._get_guide_counts_bcmatch()
        else:  # count both bc matched & unmatched guide counts
            self._get_guide_counts_bcmatch_semimatch()

        if self.count_guide_edits:
            self.screen.uns["guide_edit_counts"] = _multiindex_dict_to_df(
                self.screen.uns["guide_edit_counts"], "edit", self.database_id
            )
            self.screen.uns["guide_edit_counts"].guide = self.screen.guides.index[
                self.screen.uns["guide_edit_counts"].guide.to_numpy(dtype=int)
            ]

        if self.count_reporter_edits:
            self.screen.uns["edit_counts"] = pd.DataFrame.from_dict(self.screen.uns["edit_counts"])

        if self.count_edited_alleles:
            guides = []
            alleles = []
            counts = []
            for guide, allele_to_count in self.guide_to_allele.items():
                if len(allele_to_count.keys()) == 0: continue
                guides.extend([guide]*len(allele_to_count.keys()))
                for allele, count in allele_to_count.items():
                    alleles.append(allele)
                    counts.append(count)
            if not (len(guides) == len(alleles) == len(counts)): 
                raise ValueError("Guides:{}, alleles:{}, counts:{}".format(len(guides), len(alleles), len(counts)))
            self.screen.uns["allele_counts"] = pd.DataFrame({"guide": guides, "allele": alleles, self.database_id: counts})
            
            if 'guide' in self.screen.uns["allele_counts"].columns:
                self.screen.uns["allele_counts"].guide = self.screen.guides.index[self.screen.uns["allele_counts"].guide]

        if self.count_guide_reporter_alleles:
            guides = []
            guide_alleles = []
            reporter_alleles = []
            counts = []
            for guide, allele_to_count in self.guide_to_guide_reporter_allele.items():
                if len(allele_to_count.keys()) == 0: continue
                guides.extend([guide]*len(allele_to_count.keys()))
                for key, count in allele_to_count.items():
                    reporter_allele, guide_allele = key
                    guide_alleles.append(guide_allele)
                    reporter_alleles.append(reporter_allele)
                    counts.append(count)
            if not (len(guides) == len(reporter_alleles) == len(guide_alleles) == len(counts)): 
                raise ValueError("Guides:{}, guide_alleles:{}, reporter_alleles: {}, counts:{}".format(
                    len(guides), len(guide_alleles), len(reporter_alleles), len(counts)))
            self.screen.uns["guide_reporter_allele_counts"] = pd.DataFrame(
                {"guide": guides, "guide_allele": guide_alleles, "reporter_allele": reporter_alleles, self.database_id: counts})
            
            if 'guide' in self.screen.uns["guide_reporter_allele_counts"].columns:
                self.screen.uns["guide_reporter_allele_counts"].guide = self.screen.guides.index[self.screen.uns["guide_reporter_allele_counts"].guide]            
        
        count_stat_path = self._jp("mapping_stats.txt")
        count_stat_file = open(count_stat_path, "w")
        count_stat_file.write("Read count with \n")
        count_stat_file.write("Unique guide match without barcode:\t{}\n".format(self.semimatch))
        count_stat_file.write("Unique guide match with barcode:\t{}\n".format(self.bcmatch))
        count_stat_file.write("No match:\t{}\n".format(self.nomatch))
        count_stat_file.write("Duplicate match with barcode:\t{}\n".format(self.duplicate_match))
        count_stat_file.write("No match with barcode & Duplicate match w/o barcode:\t{}\n".format(self.duplicate_match_wo_barcode))

    def _get_guide_counts_bcmatch(self):
        NotImplemented

    def _count_guide_edits(self, matched_guide_idx, R1_record: SeqIO.SeqRecord, single_base_qual_cutoff = 30):
        strand_str_to_int = {"neg": -1, "pos": 1}
        if self.guides_has_strands:
            try:
                guide_strand = strand_str_to_int[
                    self.screen.guides.Strand[matched_guide_idx]
                ]
            except KeyError:  # control guides
                guide_strand = 1
        else:
            guide_strand = 1
        ref_guide_seq = self.screen.guides.sequence[matched_guide_idx]
        read_guide_seq, read_guide_qual = self.get_guide_seq_qual(R1_record, len(ref_guide_seq))
        guide_edit_allele = _get_edited_allele_crispresso(
            ref_seq=ref_guide_seq,
            query_seq=read_guide_seq,
            aln_mat_path = self.output_dir + "/.aln_mat.txt",
            offset=0,
            strand=guide_strand,
            start_pos=0,
            end_pos=len(ref_guide_seq),
            positionwise_quality = np.array(read_guide_qual),
            quality_thres = single_base_qual_cutoff
        )
        return(guide_edit_allele)
        

    def _count_reporter_edits(self, matched_guide_idx: int, R1_seq, R2_record: SeqIO.SeqRecord, 
    single_base_qual_cutoff=30, guide_allele = None,):
        '''
        Count edits in single read to save as allele.
        '''
        strand_str_to_int = {"neg": -1, "pos": 1}
        ref_reporter_seq = self.screen.guides.Reporter[matched_guide_idx]
        read_reporter_seq, read_reporter_qual = self.get_reporter_seq_qual(R2_record)

        if self.guides_has_strands:
            try:
                guide_strand = strand_str_to_int[
                    self.screen.guides.Strand[matched_guide_idx]
                ]
                if guide_strand == -1:
                    offset = (
                        self.screen.guides.start_pos[matched_guide_idx]
                        + 32 - 6 - 1
                    )
                if guide_strand == 1:
                    offset = self.screen.guides.start_pos[matched_guide_idx] \
                        - (32 - 6 - self.screen.guides.guide_len[matched_guide_idx])
            except KeyError:  # control guides
                guide_strand = 1
                offset = 0
        else:
            guide_strand = 1
            offset = -(self.screen.guides["Target base position in reporter"][matched_guide_idx] -1)
            # TODO: clean this up
        
        allele = _get_edited_allele_crispresso(
            ref_seq=ref_reporter_seq,
            query_seq=read_reporter_seq,
            aln_mat_path = self.output_dir + "/.aln_mat.txt",
            offset=offset,
            strand=guide_strand,
            positionwise_quality = np.array(read_reporter_qual),
            quality_thres = single_base_qual_cutoff
        )

        if allele.edits:
            if self.count_edited_alleles:
                if matched_guide_idx in self.guide_to_allele.keys():
                    if allele in self.guide_to_allele[matched_guide_idx].keys():
                        self.guide_to_allele[matched_guide_idx][allele] += 1
                    else:
                        self.guide_to_allele[matched_guide_idx][allele] = 1
                else:
                    self.guide_to_allele[matched_guide_idx] = {allele: 1}
            if self.count_guide_reporter_alleles and (not guide_allele is None):
                if matched_guide_idx in self.guide_to_guide_reporter_allele.keys():
                    if (allele, guide_allele) in self.guide_to_guide_reporter_allele[matched_guide_idx].keys():
                        self.guide_to_guide_reporter_allele[matched_guide_idx][(allele, guide_allele)] += 1
                    else:
                        self.guide_to_guide_reporter_allele[matched_guide_idx][(allele, guide_allele)] = 1
                else:
                    self.guide_to_guide_reporter_allele[matched_guide_idx] = {(allele, guide_allele): 1}
        
    def _get_guide_counts_bcmatch_semimatch(
        self, bcmatch_layer="X_bcmatch", semimatch_layer="X"
    ):

        self.screen.layers[semimatch_layer] = np.zeros_like((self.screen.X))
        R1_iter, R2_iter = self._get_fastq_iterators()

        outfile_R1_nomatch, outfile_R2_nomatch = self._get_fastq_handle("nomatch")
        outfile_R1_semimatch, outfile_R2_semimatch = self._get_fastq_handle("semimatch")
        outfile_R1_dup_wo_bc, outfile_R2_dup_wo_bc = self._get_fastq_handle(
            "duplicate_wo_barcode"
        )
        outfile_R1_dup, outfile_R2_dup = self._get_fastq_handle("duplicate")

        for i, (r1, r2) in tqdm(
            enumerate(zip(R1_iter, R2_iter)), total=self.n_reads_after_filtering
        ):
            R1_seq = str(r1.seq)
            R2_seq = str(r2.seq)

            bc_match, semimatch = self._match_read_to_sgRNA_bcmatch_semimatch(
                R1_seq, R2_seq
            )

            if len(bc_match) == 0:
                if len(semimatch) == 0:  # no guide match
                    if self.keep_intermediate:
                        _write_paired_end_reads(
                            r1, r2, outfile_R1_nomatch, outfile_R2_nomatch
                        )
                    self.nomatch += 1
                elif len(semimatch) >= 2:  # Duplicate match if w/o barcode
                    if self.keep_intermediate:
                        _write_paired_end_reads(
                            r1, r2, outfile_R1_dup_wo_bc, outfile_R2_dup_wo_bc
                        )
                    self.duplicate_match_wo_barcode += 1
                else:  # guide match with no barcode match
                    matched_guide_idx = semimatch[0]
                    try:
                        self.screen.layers[semimatch_layer][matched_guide_idx, 0] += 1
                    except:
                        print(semimatch)

                    if self.count_guide_edits:
                        self._count_guide_edits(matched_guide_idx, r1)
                    self.semimatch += 1

            elif len(bc_match) >= 2:  # duplicate mapping
                if self.keep_intermediate:
                    _write_paired_end_reads(r1, r2, outfile_R1_dup, outfile_R2_dup)
                self.duplicate_match += 1

            else:  # unique barcode match
                matched_guide_idx = bc_match[0]
                self.screen.layers[bcmatch_layer][matched_guide_idx, 0] += 1
                self.bcmatch += 1
                if self.count_guide_edits or self.count_guide_reporter_alleles:
                    guide_allele = self._count_guide_edits(matched_guide_idx, r1)
                if self.count_reporter_edits or self.count_edited_alleles or self.count_guide_reporter_alleles:
                    # TBD: what if reporter seq doesn't match barcode & guide?
                    if self.count_guide_reporter_alleles:
                        self._count_reporter_edits(matched_guide_idx, R1_seq, r2, guide_allele = guide_allele)
                    else: 
                        self._count_reporter_edits(matched_guide_idx, R1_seq, r2)

        self.screen.X = (
            self.screen.layers[semimatch_layer] + self.screen.layers[bcmatch_layer]
        )

    def _write_allele(self, guide_idx: int, allele: Allele, uns_key="allele_counts"):
        # if len(allele.edits) == 0:
        #     return

        if (guide_idx, str(allele)) in self.screen.uns[uns_key].keys():
            self.screen.uns[uns_key][(guide_idx, str(allele))] += 1
        else:
            self.screen.uns[uns_key][(guide_idx, str(allele))] = 1

    def _write_edits(self, guide_idx: int, allele: Allele, uns_key="edit_counts"):
        for edit in allele.edits:
            if (guide_idx, str(edit)) in self.screen.uns[uns_key].keys():
                self.screen.uns[uns_key][(guide_idx, str(edit))] += 1
            else:
                self.screen.uns[uns_key][(guide_idx, str(edit))] = 1

    def _write_guide_reporter_allele(self, guide_idx: int, allele: Allele, guide_allele: Allele, uns_key="guide_reporter_allele_counts"):
        if len(allele.edits) == 0 and len(guide_allele.edits) == 0:
            return
        if (guide_idx, str(allele), str(guide_allele)) in self.screen.uns[uns_key].keys():
            self.screen.uns[uns_key][(guide_idx, str(allele), str(guide_allele))] += 1
        else:
            self.screen.uns[uns_key][(guide_idx, str(allele), str(guide_allele))] = 1

    def get_guide_seq(self, R1_seq, R2_seq, guide_length):
        """This can be edited by user based on the read construct."""
        #_seq_match = np.where(seq.replace(self.base_edited_from, self.base_edited_to) == self.screen.guides.masked_sequence)[0]
        guide_start_idx = R1_seq.replace(self.base_edited_from, self.base_edited_to).find(
            self.guide_start_seq.replace(self.base_edited_from, self.base_edited_to))
        if guide_start_idx == -1:
            return None
        if guide_start_idx + guide_length >= len(R1_seq): 
            return None
        guide_start_idx = guide_start_idx + len(self.guide_start_seq)
        gRNA_seq = R1_seq[guide_start_idx : (guide_start_idx + guide_length)]
        if len(gRNA_seq) != guide_length:
            return None
        return gRNA_seq

    def get_guide_seq_qual(self, R1_record: SeqIO.SeqRecord, guide_length):
        R1_seq = R1_record.seq
        guide_start_idx = R1_seq.replace(self.base_edited_from, self.base_edited_to).find(
            self.guide_start_seq.replace(self.base_edited_from, self.base_edited_to))
        if guide_start_idx == -1:
            return None, None
        if guide_start_idx + guide_length >= len(R1_seq): 
            return None, None
        guide_start_idx = guide_start_idx + len(self.guide_start_seq)
        seq = R1_record[guide_start_idx : (guide_start_idx + guide_length)]
        return(str(seq.seq), seq.letter_annotations["phred_quality"])

    def get_reporter_seq(self, R1_seq, R2_seq):
        """This can be edited by user based on the read construct."""
        reporter_seq = revcomp(
            R2_seq[self.guide_bc_len : (self.guide_bc_len + self.reporter_length)]
        )
        return(reporter_seq)

    def get_reporter_seq_qual(self, R2_record: SeqIO.SeqRecord):
        seq = R2_record[self.guide_bc_len : (self.guide_bc_len + self.reporter_length)].reverse_complement()
        return(str(seq.seq), seq.letter_annotations["phred_quality"])

    def get_barcode(self, R1_seq, R2_seq):
        """This can be edited by user based on the read construct."""
        return revcomp(R2_seq[: self.guide_bc_len])

    def _match_read_to_sgRNA_bcmatch_semimatch(self, R1_seq: str, R2_seq: str):
        # This should be adjusted for each experimental recipes.'
        guide_barcode = self.get_barcode(R1_seq, R2_seq)
        bc_match_idx = np.array([])
        semimatch_idx = np.array([])
        for guide_length in self.guide_lengths:
            seq = self.get_guide_seq(R1_seq, R2_seq, guide_length)
            if seq is None:
                continue

            _seq_match = np.where(seq.replace(self.base_edited_from, self.base_edited_to) == self.screen.guides.masked_sequence)[0]
            _bc_match = np.where(guide_barcode.replace(self.base_edited_from, self.base_edited_to) == self.screen.guides.masked_barcode)[0]
            
            bc_match_idx = np.append(bc_match_idx, np.intersect1d(_seq_match, _bc_match))
            semimatch_idx = np.append(semimatch_idx, _seq_match)

        return (bc_match_idx.astype(int), semimatch_idx.astype(int))

    def _get_guide_position_seq_of_read(self, seq):
        guide_start_idx = self._get_guide_start_idx(seq)
        if guide_start_idx == -1:
            return None

        return [
            seq[guide_start_idx : (guide_start_idx + guide_length)]
            for guide_length in self.guide_lengths
        ]

    def _get_guide_start_idx(self, seq):
        start_seq_idx = seq.find(self.guide_start_seq)
        if start_seq_idx == -1:
            return -1
        return start_seq_idx + len(self.guide_start_seq)

    def get_gRNA_barcode(self, R1_seq, R2_seq):
        # This can be adjusted for different construct design.
        return revcomp(R2_seq[: self.guide_bc_len])

    def _get_fastq_handle(
        self,
        out_type: str = None,
    ):
        assert out_type in [
            "semimatch", "nomatch", "duplicate_wo_barcode", "duplicate"
        ]
        R1_filename = self.R1_filename.replace(".fastq", "_{}.fastq".format(out_type))
        R2_filename = self.R2_filename.replace(".fastq", "_{}.fastq".format(out_type))
        R1_handle = _get_fastq_handle(R1_filename, "w")
        R2_handle = _get_fastq_handle(R2_filename, "w")

        return (R1_handle, R2_handle)

    def _get_fastq_iterators(self):
        R1_handle = _get_fastq_handle(self.R1_filename)
        R2_handle = _get_fastq_handle(self.R2_filename)

        R1_iterator = FastqPhredIterator(R1_handle)
        R2_iterator = FastqPhredIterator(R2_handle)

        return (R1_iterator, R2_iterator)

    def _get_seq_records(self):
        R1_handle = _get_fastq_handle(self.R1_filename)
        R2_handle = _get_fastq_handle(self.R2_filename)
        R1 = list(SeqIO.parse(R1_handle, "fastq"))
        R2 = list(SeqIO.parse(R2_handle, "fastq"))
        R1_handle.close()
        R2_handle.close()
        return (R1, R2)

    def _check_names_filter_fastq(self, filter_by_qual=False):
        if self.min_average_read_quality > 0 or self.min_single_bp_quality > 0:
            info(
                "Filtering reads with average bp quality < {} and single bp quality < {} ...".format(
                    self.min_average_read_quality, self.min_single_bp_quality
                )
            )

        if self.qend_R1 > 0 or self.qend_R2 > 0:
            info(
                "In the filtering, bases up to position {} of R1 and {} of R2 are considered.".format(
                    self.qend_R1, self.qend_R2
                )
            )

        R1, R2 = self._get_seq_records()

        _check_readname_match(R1, R2)
        if filter_by_qual:
            self.n_reads_after_filtering = self._filter_read_quality(R1, R2)
            if self.n_reads_after_filtering == 0:
                raise NoReadsAfterQualityFiltering(
                    "No reads in input or no reads survived the average or single bp quality filtering."
                )
            else:
                info(
                    "Number of reads in input:%d\tNumber of reads after filtering:%d\n"
                    % (self.n_total_reads, self.n_reads_after_filtering)
                )
        else:
            self.n_reads_after_filtering = self.n_total_reads

    def _filter_read_quality(self, R1=None, R2=None) -> int:
        R1_filtered = gzip.open(self.filtered_R1_filename, "w+")
        R2_filtered = gzip.open(self.filtered_R2_filename, "w+")

        if R1 is None or R2 is None:
            R1, R2 = self._get_seq_records()

        n_reads_after_filtering = 0
        for i in range(len(R1)):
            R1_record = R1[i]
            R2_record = R2[i]

            if self.filter_by_qual:
                R1_quality_pass = _read_is_good_quality(
                    R1_record,
                    self.min_average_read_quality,
                    self.min_single_bp_quality,
                    self.qend_R1,
                )
                R2_quality_pass = _read_is_good_quality(
                    R2_record,
                    self.min_average_read_quality,
                    self.min_single_bp_quality,
                    self.qend_R2,
                )

                if R1_quality_pass and R2_quality_pass:
                    n_reads_after_filtering += 1
                    R1_filtered.write(R1.format("fastq"))
                    R2_filtered.write(R2.format("fastq"))
        return n_reads_after_filtering

    def _write_start_log(self):
        try:
            os.makedirs(self.output_dir)
            print("Creating Folder %s" % self.output_dir)
        except:
            print("Folder %s already exists." % self.output_dir)
        self.log_filename = self._jp("beretCount_RUNNING_LOG.txt")
        logging.getLogger().addHandler(logging.FileHandler(self.log_filename))
        with open(self.log_filename, "w+") as outfile:
            outfile.write(
                "[Command used]:\nCRISPRessoCount %s\n\n[Execution log]:\n"
                % " ".join(sys.argv)
            )

    def _jp(self, filename):
        return os.path.join(self.output_dir, filename)

    def _get_database_name(self):
        get_name_from_fasta = (
            lambda x: os.path.basename(x)
            .replace(".fastq", "")
            .replace(".gz", "")
            .replace("_R1", "")
        )

        if not self.name:
            database_id = "%s" % get_name_from_fasta(self.R1_filename)
        else:
            database_id = self.name
        return database_id



import pandas as pd
import seaborn as sns
import matplotlib
import matplotlib.pyplot as plt
from scipy import stats
import numpy as np


def collect_guide_edit_counts(reps, conditions, guide_filename_pattern, reporter_filename_pattern, guide_info_filename, 
                             ref_base = "A", sample_base = "G"):
    counts_df = pd.read_csv(guide_info_filename)
    counts_df = counts_df.set_index("name")
    
    for rep in reps:
        for cond in conditions:
            res = pd.read_csv(guide_filename_pattern.format(r = rep, c = cond), delimiter="\t")
            res = res.set_index("guide_name")
            mapped_reads = res["read_counts"].sum()
            res["guide_RPM"] = res["read_counts"] / mapped_reads * 10**6
            res.columns = res.columns.map(lambda x: "{r}_{c}_".format(r = rep, c = cond) + x)
            assert all([x in counts_df.index.tolist() for x in res.index.tolist()])
            cols_to_join = res.columns[[0, 2, 3]]
            counts_df = counts_df.join(res[cols_to_join], how = "left")

            res = pd.read_csv(reporter_filename_pattern.format(r = rep, c = cond), delimiter="\t")
            res = res.set_index("name")
            all_edits = res.groupby('name')["count"].sum()
            all_edits.name = "{r}_{c}_all_edits".format(r = rep, c = cond)
            abe_edits = res.loc[(res.ref_base == ref_base) & (res.sample_base == sample_base),:].groupby('name')["count"].sum()
            abe_edits.name = "{r}_{c}_AG_edits".format(r = rep, c = cond)
            
            # count exact match in position
            all_reporter_edits_df = counts_df.join(res, on = "name", how = "left")
            all_reporter_edits_df = all_reporter_edits_df.reset_index()
            exact_edits = all_reporter_edits_df.loc[(all_reporter_edits_df["Target base position in gRNA"] == (all_reporter_edits_df.pos + 1)) & (all_reporter_edits_df.ref_base == ref_base) & (all_reporter_edits_df.sample_base == sample_base)][["name", "count"]].set_index("name", drop = True)["count"]
            exact_edits.name = "{r}_{c}_exact_edits".format(r = rep, c = cond)

            counts_df = counts_df.join(all_edits, how = "left")
            counts_df = counts_df.join(abe_edits, how = "left")
            counts_df = counts_df.join(exact_edits, how = "left")
            
            counts_df["{r}_{c}_edit_RPM".format(r = rep, c = cond)] = counts_df["{r}_{c}_exact_edits".format(r = rep, c = cond)] / mapped_reads * 10**6
            counts_df["{r}_{c}_edit_rate".format(r = rep, c = cond)] = counts_df["{r}_{c}_exact_edits".format(r = rep, c = cond)] / counts_df["{r}_{c}_read_counts".format(r = rep, c = cond)]
        
    counts_df = counts_df.fillna(0)
    return(counts_df)


def write_count_stats(count_file_name, reps, outfile_name, condition = "bulk", stat_fn = np.median):
    counts_df = pd.read_csv(count_file_name)
    edit_rates = counts_df.set_index("name")[["{r}_bulk_edit_rate".format(r = r) for r in reps]]
    edit_rates["edit_rate"] = edit_rates.apply(stat_fn, axis = 1)
    guide_counts = counts_df.set_index("name")[["{r}_bulk_read_counts".format(r = r) for r in reps]]
    guide_counts["guide_count"] = guide_counts.apply(stat_fn, axis = 1)
    edit_rates[["edit_rate"]].join(guide_counts[["guide_count"]]).to_csv(outfile_name)
    

def _log2FC(x, y):
    with np.errstate(divide='ignore', invalid='ignore'):
        return(np.log2((x+0.5)/(y+0.5)))

    
def _get_log2FC(df, prefix, cond1, cond2, measure, count_thres = 5, count_column = "read_counts"):
    res = _log2FC(df["{p}_{c}_{m}".format(p = prefix, c = cond1, m = measure)], 
                df["{p}_{c}_{m}".format(p = prefix, c = cond2, m = measure)])
    ignore_idx = np.where(np.logical_and(df["{p}_{c}_{cn}".format(p = prefix, c = cond1, cn = count_column)] < count_thres,
                         df["{p}_{c}_{cn}".format(p = prefix, c = cond2, cn = count_column)] < count_thres))
    res.iloc[ignore_idx] = np.nan
    return(res)
    
    
def get_log2FC(count_file_name, reps, cond1 = "top", cond2 = "bot", guide_count_measure = "RPM", 
              edit_count_measure = "edit_RPM", guide_count_filter_measure = "read_counts", 
              edit_count_filter_measure = "exact_edits"):
    
    counts_df = pd.read_csv(count_file_name)
    guide_enrichment = counts_df[["name"]]
    
    for rep in reps:
        top_vs_bulk = _get_log2FC(counts_df, rep, "top", "bulk", guide_count_measure, count_column = "read_counts")
        bot_vs_bulk = _get_log2FC(counts_df, rep, "bot", "bulk", guide_count_measure, count_column = "read_counts")
        bot_vs_top = _get_log2FC(counts_df, rep, "bot", "top", guide_count_measure, count_column = "read_counts")
        top_vs_bulk.name = rep + "_guide_top_vs_bulk"
        bot_vs_bulk.name = rep + "_guide_bot_vs_bulk"
        bot_vs_top.name = rep + "_guide_bot_vs_top"

        guide_enrichment = pd.concat((guide_enrichment, top_vs_bulk, bot_vs_bulk, bot_vs_top), axis = 1)

        # Reporter enrichment
        top_vs_bulk = _get_log2FC(counts_df, rep, "top", "bulk", "edit_RPM", count_column = edit_count_filter_measure)
        bot_vs_bulk = _get_log2FC(counts_df, rep, "bot", "bulk", "edit_RPM", count_column = edit_count_filter_measure)
        bot_vs_top = _get_log2FC(counts_df, rep, "bot", "top", "edit_RPM", count_column = edit_count_filter_measure)
        top_vs_bulk.name = rep + "_edit_top_vs_bulk"
        bot_vs_bulk.name = rep + "_edit_bot_vs_bulk"
        bot_vs_top.name = rep + "_edit_bot_vs_top"

        guide_enrichment = pd.concat((guide_enrichment, top_vs_bulk, bot_vs_bulk, bot_vs_top), axis = 1)
    return(guide_enrichment)
        
    

def plot_count_dist_ranked(count_file_name, reps, conditions, sample = "",
    figsize = None):
    if figsize is None:
        figsize = (3*len(conditions), 9)
    plt.rcParams.update({'font.size': 15})

    counts_df = pd.read_csv(count_file_name)

    figs, axs = plt.subplots(3, len(conditions), figsize=figsize)
    for j, c in enumerate(conditions):
        counts_df["RPM_sum"] = 0

        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            counts_df.RPM_sum += counts_df[prefix + "RPM"]
        counts_df = counts_df.sort_values("RPM_sum", ascending = False)
        counts_df = counts_df.reset_index(drop = True)

        # Panel 1: guide counts
        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            p = axs[0, j].plot(counts_df.index, counts_df[prefix + "read_counts"], ".", label = r, markersize = 0.5, alpha = 0.5)
        #lgnd = axs[0, j].legend()

        axs[0, j].set_title(c)
        axs[0, j].set_ylabel("guide counts")
        axs[0, j].set_xlabel("rank by guide count")

            # Panel 2: edit rates sorted by guide counts
        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            axs[1, j].plot(counts_df.index, counts_df[prefix + "edit_rate"], ".", markersize = 0.5, alpha = 0.5, label =r)
        #axs[1, j].set_title("LDLRCDS edit efficiency, sorted by total guide RPM ({})".format(c))
        axs[1, j].set_ylabel("edit rate")
        axs[1, j].set_xlabel("guide rank")
        axs[1, j].set_xlabel("rank by guide count")

        # Panel 3: Edit rates
        counts_df["mean_edit_rate"] = 0
        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            counts_df.mean_edit_rate += counts_df[prefix + "exact_edits"]
        counts_df.mean_edit_rate /= 4
        counts_df = counts_df.sort_values("mean_edit_rate", ascending = False)
        counts_df = counts_df.reset_index(drop = True)
        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            axs[2, j].plot(counts_df.index, counts_df[prefix + "exact_edits"], ".", markersize = 0.5, alpha = 0.5, label = r)
        #axs[2, j].set_title("LDLRCDS edit efficiency, sorted by mean edit rate ({})".format(c))
        axs[2, j].set_ylabel("exact edit counts")
        axs[2, j].set_xlabel("rank by edit count")

    if len(reps) > 1:
        lgnd = axs[1, len(conditions)-1].legend(bbox_to_anchor=(1.04,0.5), loc="center left", borderaxespad=0)
        for i in range(len(reps)):
            lgnd.legendHandles[i]._legmarker.set_markersize(6)
    plt.suptitle("{} guide / edit counts".format(sample))
    plt.tight_layout()

    
def plot_count_dist(count_file_name, reps, conditions, sample = "",
    figsize = None):
    if figsize is None:
        figsize = (3*len(conditions), 9)
    plt.rcParams.update({'font.size': 15})
    counts_df = pd.read_csv(count_file_name)

    figs, axs = plt.subplots(3, len(conditions), figsize=figsize)
    for j, c in enumerate(conditions):
        counts_df["RPM_sum"] = 0

        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            counts_df.RPM_sum += counts_df[prefix + "RPM"]
        
        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            sns.kdeplot(ax = axs[0, j], x = counts_df[prefix + "read_counts"], label = r, alpha = 0.5).set(
                xlim=(0, max(counts_df[prefix + "read_counts"])))
        #lgnd = axs[0, j].legend()

        axs[0, j].set_title(c)
        axs[0, j].set_xlabel("guide counts")
        axs[0, 0].set_ylabel("Density")

            # Panel 2: edit rates sorted by guide counts
        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            sns.kdeplot(ax = axs[1, j], x = counts_df[prefix + "edit_rate"], alpha = 0.5, label =r).set(xlim=(0, 1))
        #axs[1, j].set_title("LDLRCDS edit efficiency, sorted by total guide RPM ({})".format(c))
        axs[1, j].set_xlabel("edit rate")
        axs[1, 0].set_ylabel("Density")

        # Panel 3: Edit rates
        counts_df["mean_edit_rate"] = 0
        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            counts_df.mean_edit_rate += counts_df[prefix + "exact_edits"]
        counts_df.mean_edit_rate /= 4
        counts_df = counts_df.sort_values("mean_edit_rate", ascending = False)
        counts_df = counts_df.reset_index(drop = True)
        for r in reps:
            prefix = "{r}_{c}_".format(r = r, c = c)
            sns.kdeplot(ax = axs[2, j], x = counts_df[prefix + "exact_edits"], alpha = 0.5, label = r).set(
                xlim=(0, max(counts_df[prefix + "exact_edits"])))
        axs[2, j].set_xlabel("exact edit counts")
        axs[2, 0].set_ylabel("Density")

    if len(reps) > 1:
        lgnd = axs[1, len(conditions)-1].legend(bbox_to_anchor=(1.04,0.5), 
            loc="center left", borderaxespad=0)

    plt.suptitle("{} count / edit rate distribution".format(sample))
    plt.tight_layout()

    
def plot_counts_vs_edit_rate(count_file_name, reps, conditions, 
    title = "Guide counts vs edit rate", outfile = None, figsize = None):
    if figsize is None:
        figsize = (3*len(conditions), 3)
    counts_df = pd.read_csv(count_file_name)
    fig, axs = plt.subplots(1, len(conditions), figsize = figsize)
    for j, c in enumerate(conditions):
        read_counts = pd.Series(dtype = 'float')
        edit_rates = pd.Series(dtype = 'float')
        for r in reps:
            prefix = "{r}_{c}".format(r = r, c = c)
            axs[j].plot(counts_df[prefix + "_read_counts"], counts_df[prefix + "_edit_rate"], ".", 
                label = prefix, markersize = 1, alpha = 0.1)
            read_counts = pd.concat((read_counts, counts_df[prefix + "_read_counts"]))
            edit_rates = pd.concat((edit_rates, counts_df[prefix + "_edit_rate"]))
        cor = round(read_counts.corr(edit_rates, "spearman"), 2)
        axs[j].text(max(read_counts)*0.4, max(edit_rates)*0.7, "cor_Sp={}".format(cor), fontsize = 10)
        axs[j].set_title(c)
        axs[j].set_ylabel("edit rate")
        axs[j].set_xlabel("guide counts")
    
    if len(reps) > 1:
        lgnd = axs[len(conditions)-1].legend(
            bbox_to_anchor=(1.04,0.45), 
            loc="center left", 
            borderaxespad=0)
        for i in range(len(lgnd.legendHandles)):
            lgnd.legendHandles[i]._legmarker.set_markersize(6)
            lgnd.legendHandles[i]._legmarker.set_alpha(1)
        
    plt.suptitle(title)
    plt.tight_layout()
    plt.show()
    if outfile: plt.savefig(outfile)
        

def _plot_replicate_pairplot(df, reps, measure = "read_counts"):
    count_df = df.iloc[:, np.append(np.where(df.columns == "name")[0], np.where(df.columns.map(lambda x: measure in x))[0])] # select columns with measure in their name
    count_df = pd.wide_to_long(count_df, reps, "name", "condition", suffix = r"(.*_{})".format(measure), sep = "_")
    count_df = count_df.reset_index()
    return(sns.pairplot(count_df, kind = "scatter", hue = "condition", plot_kws={"s": 0.5}))


def _plot_replicate_pairplot_foldchange(df, reps, measure = "guide_bot_vs_top"):
    count_df = df.iloc[:, np.append(np.where(df.columns == "name")[0], np.where(df.columns.map(lambda x: measure in x))[0])]
    count_df.columns = count_df.columns.map(lambda s: s.split("_" + measure)[0])
    count_df = count_df.set_index("name", drop = True)
    assert len(count_df.columns) == len(reps), "{}".format(count_df.columns)
    return(sns.pairplot(count_df, kind = "scatter"))


def _annotate_pearsonr(x, y, **kws):
    r, _ = stats.pearsonr(x, y)
    ax = plt.gca()
    # count how many annotations are already present
    n = len([c for c in ax.get_children() if isinstance(c, matplotlib.text.Annotation)])
    pos = (.1, .9 - .1*n)
    ax.annotate("{}:r={:.2f}".format(kws['label'].split("_")[0],r),
                xy=pos, xycoords=ax.transAxes)

    
def _annotate_pearsonr_foldchange(x, y, **kws):
    ignore_idx = np.isnan(x) | np.isnan(y)
    x = x[~ignore_idx]
    y = y[~ignore_idx]
    r, _ = stats.pearsonr(x, y)
    ax = plt.gca()
    
    # count how many annotations are already present
    n = len([c for c in ax.get_children() if isinstance(c, matplotlib.text.Annotation)])
    pos = (.1, .9 - .1*n)
    
    ax.annotate("r={:.2f}".format(r),
                xy=pos, xycoords=ax.transAxes)
    
    
def plot_replicate_read_count_correlation(reporter_file_name: str, reps, outfile_name = None, title = ""):
    read_count_df = pd.read_csv(reporter_file_name)
    pairplot_obj = _plot_replicate_pairplot(read_count_df, measure = "read_counts", reps = reps)
    pairplot_obj.map_lower(_annotate_pearsonr)
    #pairplot_obj.fig.suptitle(title)
    #pairplot_obj.fig.tight_layout()
    if outfile_name:
        plt.savefig(outfile_name)
        

def plot_replicate_enrichment_correlation(enrichment_df, count, reps, outfile_name = None, title = ""):
    assert count in ["edit", "guide"], "argument 'count' must be one of 'guide' or 'edit'."
    pairplot_obj = _plot_replicate_pairplot_foldchange(enrichment_df, measure = "{}_bot_vs_top".format(count), reps = reps)
    pairplot_obj.map_lower(_annotate_pearsonr_foldchange)
    pairplot_obj.fig.suptitle(title)
    pairplot_obj.fig.tight_layout()
    


from typing import List, Union, Collection
from functools import reduce
from copy import deepcopy
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from anndata import AnnData
import anndata as ad
from perturb_tools import Screen
from beret.annotate.AminoAcidEdit import AminoAcidAllele, CodingNoncodingAllele
from .Edit import Edit, Allele
from ..framework._supporting_fn import get_aa_alleles, filter_allele_by_pos


def _get_counts(filename_pattern, guides, reps, conditions):
    for rep in reps:
        for cond in conditions:
            res = pd.read_csv(filename_pattern.format(r=rep, c=cond), delimiter="\t")
            res = res.set_index("guide_name")[["read_counts"]]
            res.columns = res.columns.map(lambda x: "{r}_{c}".format(r=rep, c=cond))
            guides = guides.join(res, how="left")
        else:
            guides = guides.fillna(0)
            return guides


def _get_edits(filename_pattern, guide_info, reps, conditions, count_exact=True):
    edits = pd.DataFrame(index=(guide_info.index))
    for rep in reps:
        for cond in conditions:
            res = pd.read_csv(filename_pattern.format(r=rep, c=cond), delimiter="\t")
            res = res.set_index("name")
            if count_exact:
                res_info = guide_info.join(res, how="left").reset_index()
                this_edit = res_info.loc[
                    (
                        (res_info["Target base position in gRNA"] == res_info.pos + 1)
                        & (res_info.ref_base == "A")
                        & (res_info.sample_base == "G")
                    )
                ][["name", "count"]].set_index("name", drop=True)["count"]
            else:
                this_edit = (
                    res.loc[(res.ref_base == "A") & (res.sample_base == "G"), :]
                    .groupby("name")["count"]
                    .sum()
                )
            this_edit.name = "{r}_{c}".format(r=rep, c=cond)
            edits = edits.join(this_edit, how="left")
        else:
            edits = edits.fillna(0)
            return edits


class ReporterScreen(Screen):
    def __init__(self, X=None, X_edit=None, X_bcmatch=None, *args, **kwargs):
        (super().__init__)(X, *args, **kwargs)
        if (X_edit is None) != (X_bcmatch is None):
            raise ValueError("Only one of number of edits or barcode matched guide counts is specified.")
        if not X_edit is None:
            self.layers["edits"] = X_edit
        if not X_bcmatch is None:
            self.layers["X_bcmatch"] = X_bcmatch
        for k, df in self.uns.items():
            if "guide" in df.columns:
                if "allele" in df.columns: 
                    self.uns[k].allele = self.uns[k].allele.map(lambda s: Allele.from_str(s))
                if "edit" in df.columns:
                    self.uns[k].edit = self.uns[k].edit.map(lambda s: Edit.from_str(s))
                if 'reporter_allele' in df.columns and 'guide_allele' in df.columns:
                    self.uns[k].reporter_allele = self.uns[k].reporter_allele.map(lambda s: Allele.from_str(s))
                    self.uns[k].guide_allele = self.uns[k].guide_allele.map(lambda s: Allele.from_str(s))
                if "aa_allele" in df.columns:
                    self.uns[k].aa_allele = self.uns[k].aa_allele.map(lambda s: CodingNoncodingAllele.from_str(s))

        
        # if "edit_counts" in self.uns.keys():
        #     if "edit" in self.uns["edit_counts"].columns:
        #         self.uns["edit_counts"].edit = self.uns["edit_counts"].edit.map(lambda s: Edit.from_str(s))
        # if "allele_counts" in self.uns.keys():
        #     if "allele" in self.uns["allele_counts"].columns:
        #         self.uns["allele_counts"].allele = self.uns["allele_counts"].allele.map(lambda s: Allele.from_str(s))
        # if "guide_reporter_allele_counts" in self.uns.keys():
        #     self.uns["guide_reporter_allele_counts"].reporter_allele = \
        #         self.uns["guide_reporter_allele_counts"].reporter_allele.map(lambda s: Allele.from_str(s))            
        #     self.uns["guide_reporter_allele_counts"].guide_allele = \
        #         self.uns["guide_reporter_allele_counts"].guide_allele.map(lambda s: Allele.from_str(s))
        
    @classmethod
    def from_file_paths(
        cls,
        reps: List[str] = None,
        conditions: List[str] = None,
        guide_info_file_name: str = None,
        guide_count_filenames: str = None,
        guide_bcmatched_count_filenames: str = None,
        edit_count_filenames: str = None,
    ):
        guide_info = pd.read_csv(guide_info_file_name).set_index("name")
        guides = pd.DataFrame(index=(pd.read_csv(guide_info_file_name)["name"]))
        guides_lenient = _get_counts(guide_count_filenames, guides, reps, conditions)
        edits_ag = _get_edits(
            edit_count_filenames, guide_info, reps, conditions, count_exact=False
        )
        edits_exact = _get_edits(edit_count_filenames, guide_info, reps, conditions)
        if guide_bcmatched_count_filenames is not None:
            guides_bcmatch = _get_counts(
                guide_bcmatched_count_filenames, guides, reps, conditions
            )
            repscreen = cls(
                guides_lenient, edits_exact, X_bcmatch=guides_bcmatch, guides=guide_info
            )
        else:
            repscreen = cls(guides_lenient, edits_exact, guides=guide_info)
        repscreen.layers["edits_ag"] = edits_ag
        repscreen.condit["replicate"] = np.repeat(reps, len(conditions))
        repscreen.condit["sort"] = np.tile(conditions, len(reps))
        repscreen.uns["replicates"] = reps
        repscreen.uns["conditions"] = conditions
        return repscreen

    @classmethod
    def from_adata(cls, adata):
        if "X_bcmatch" in adata.layers:
            X_bcmatch = adata.layers["X_bcmatch"]
        else:
            X_bcmatch = None
        repscreen = cls(
            (adata.X),
            (adata.layers["edits"]),
            X_bcmatch=X_bcmatch,
            guides=(adata.obs),
            condit=(adata.var),
            uns=(adata.uns),
            layers=(adata.layers)
        )
        return repscreen

    def copy(self):
        adata = super().copy()
        return(type(self).from_adata(adata))

    def __add__(self, other):
        if all(self.guides.index == other.guides.index):
            if all(self.condit.index == other.condit.index):
                added_uns = dict()
                for k in self.uns.keys():
                    if k not in other.uns.keys(): continue
                    if k == "guide_edit_counts":
                        index_pair = ["guide", "edit"]
                    elif k == "edit_counts":
                        index_pair = ["guide", "edit"]
                    elif k == "allele_counts":
                        index_pair = ["guide", "allele"]
                    elif k == "guide_reporter_allele_counts":
                        index_pair = ["guide", "reporter_allele", "guide_allele"]
                    else:
                        continue
                    self_df = self.uns[k].set_index(index_pair, drop = True)
                    add_df = other.uns[k].set_index(index_pair, drop = True)
                    add_df = add_df.loc[:,self_df.columns]
                    added_uns[k] = self_df.add(add_df, fill_value = 0).astype(int).reset_index()

                if "X_bcmatch" in self.layers and "X_bcmatch" in other.layers:
                
                    added = ReporterScreen(
                        (self.X + other.X),
                        (self.layers["edits"] + other.layers["edits"]),
                        (self.layers["X_bcmatch"] + other.layers["X_bcmatch"]),
                        guides=(self.guides),
                        condit=(self.condit),
                        uns = added_uns
                    )
                else:
                    added = ReporterScreen(
                        (self.X + other.X),
                        (self.layers["edits"] + self.layers["edits"]),
                        guides=(self.guides),
                        condit=(self.condit),
                        uns = added_uns
                    )
                return added
        raise ValueError("Guides/sample description mismatch")

    def __getitem__(self, index):
        ''' TODO: currently the condit names are in ['index'] column. Making it to be the idnex will 
        allow the subsetting by condition names.
        '''
        oidx, vidx = self._normalize_indices(index)
        if isinstance(oidx, (int, np.integer)):
            if not (-self.n_obs <= oidx < self.n_obs):
                raise IndexError(f"Observation index `{oidx}` is out of range.")
            oidx += self.n_obs * (oidx < 0)
            oidx = slice(oidx, oidx + 1, 1)
        if isinstance(vidx, (int, np.integer)):
            if not (-self.n_vars <= vidx < self.n_vars):
                raise IndexError(f"Variable index `{vidx}` is out of range.")
            vidx += self.n_vars * (vidx < 0)
            vidx = slice(vidx, vidx + 1, 1)
        
        guides_include = self.guides.iloc[oidx]["name"]
        condit_include = self.condit.iloc[vidx]['index'].tolist()
        print(condit_include)
        adata = super().__getitem__(index)
        new_uns = deepcopy(adata.uns)
        for k, df in adata.uns.items():
            if "guide" in df.columns:
                if "allele" in df.columns: key_col = ["guide", "allele"]
                elif "edit" in df.columns: key_col = ["guide", "edit"]
                elif "aa_allele" in df.columns: key_col = ["guide", "aa_allele"]
                else: continue
                df_new = df.loc[df.guide.isin(guides_include), key_col + condit_include]
                df_new = df_new.loc[df_new.loc[:, condit_include].sum(axis=1) > 0, :]
                new_uns[k] = df_new
        adata.uns = new_uns
        return(type(self).from_adata(adata))



    def get_edit_mat_from_uns(
        self, 
        ref_base, 
        alt_base, 
        match_target_position=True,
        rel_pos_start = 0,
        rel_pos_end = np.Inf,
        rel_pos_is_reporter = True,
        target_pos_col = "target_pos"
        ):
        edits = self.uns["edit_counts"].copy()
        if not self.layers["edits"] is None:
            old_edits = self.layers["edits"].copy()
            self.layers["edits"] = np.zeros_like(
                self.X,
            )
        else:
            old_edits = None
        edits["ref_base"] = edits.edit.map(lambda e: e.ref_base)
        edits["alt_base"] = edits.edit.map(lambda e: e.alt_base)
        edits = edits.loc[(edits.ref_base == ref_base) & (edits.alt_base == alt_base),:].reset_index()
        guide_len = self.guides.sequence.map(len)
        guide_name_to_idx = self.guides[["name"]].reset_index().set_index("name")
        edits["guide_idx"] = guide_name_to_idx.loc[edits.guide, "index"].reset_index(drop=True)
        edits["guide_start_pos"] = 32 - 6 - guide_len[edits.guide_idx].reset_index(drop=True)
        if not match_target_position:
            edits["rel_pos"] = edits.edit.map(lambda e: e.rel_pos)
            edits["in_rel_pos_range"] = (edits.rel_pos >= rel_pos_start + edits.guide_start_pos) & (edits.rel_pos < rel_pos_end + edits.guide_start_pos)
            good_edits = edits.loc[edits.in_rel_pos_range, ["guide", "edit"] + self.condit["index"].tolist()]
        else:
            edits["rel_pos"] = edits.edit.map(lambda e: e.rel_pos)
            edits["target_pos_guides"] = self.guides.loc[edits.guide_idx, target_pos_col].reset_index(drop=True)
            edits["target_pos_matches"] = edits.rel_pos == edits.target_pos_guides
            good_edits = edits.loc[edits.target_pos_matches, ["guide", "edit"] + self.condit["index"].tolist()]
        
        good_guide_idx = guide_name_to_idx.loc[good_edits.guide, "index"].astype(int)
        try:
            for gidx, eidx in zip(good_guide_idx, good_edits.index):
                self.layers["edits"][gidx, :] += good_edits.loc[eidx, self.condit["index"].tolist()].astype(int)
        except:
            return((good_guide_idx, good_edits))
        print("New edit matrix saved in .layers['edits']. Returning old edits.")
        return old_edits

    def get_edit_rate(
        self, 
        normalize_by_editable_base = False, 
        edited_base = None,
        editable_base_start = 3,
        editable_base_end = 8,
        bcmatch_thres = 1,
        prior_weight: float = None,
        return_result = False,
        count_layer = "X_bcmatch",
        edit_layer = "edits"
        ):
        """
        prior_weight: 
        Considering the edit rate to have prior of beta distribution with mean 0.5,
        prior weight to use when calculating posterior edit rate.
        """
        if self.layers[count_layer] is not None and self.layers[edit_layer] is not None:
            num_targetable_sites = 1.0
            if normalize_by_editable_base:
                if not edited_base in ["A", "C", "T", "G"]: 
                    raise ValueError("Specify the correct edited_base")
                num_targetable_sites = self.guides.sequence.map(
                    lambda s: s[editable_base_start:editable_base_end].count(edited_base))
            bulk_idx = np.where(
                self.condit.reset_index()["index"].map(lambda s: "bulk" in s)
            )[0]

            if prior_weight is None:
                prior_weight = 1
            n_edits = self.layers[edit_layer][:, bulk_idx].sum(axis=1)
            n_counts = self.layers[count_layer][:, bulk_idx].sum(axis=1)
            self.guides["edit_rate"] = (n_edits + prior_weight / 2) / \
                ((n_counts*num_targetable_sites) + prior_weight/2)
            self.guides["edit_rate"][n_counts < bcmatch_thres] = np.nan
            if normalize_by_editable_base:
                self.guides["edit_rate"][num_targetable_sites == 0] = np.nan
            if return_result:
                return(n_edits, n_counts)
        else:
            raise ValueError("edits or barcode matched guide counts not available.")

    def get_edit_from_allele(self):
        df = self.uns["allele_counts"].copy()
        df["edits"] = df.allele.map(lambda a: str(a).split(","))
        df = df.explode("edits").groupby(["guide", "edits"]).sum()
        df = df.reset_index().rename(columns={"edits":"edit"})
        df["edit"] = df.edit.map(lambda s: Edit.from_str(s))
        return(df)

    def _get_allele_norm(self, allele_count_df = None, thres = 10):
        '''
        Get bcmatched count to normalize allele counts
        thres: normalizing count below this number would be disregarded.
        '''
        if allele_count_df is None:
            allele_count_df = self.uns["allele_counts"]
        guide_to_idx = self.guides[["name"]].reset_index().set_index("name")
        guide_idx = guide_to_idx["index"][allele_count_df.guide]
        norm_counts = self.layers["X_bcmatch"][guide_idx.values.astype(int), :]
        norm_counts[norm_counts < thres] = np.nan
        return(norm_counts)

    def get_normalized_allele_counts(self, allele_count_df = None, norm_thres = 10):
        if allele_count_df is None:
            allele_count_df = self.uns["allele_counts"]
        alleles_norm = allele_count_df.copy()
        count_columns = self.condit["index"]
        norm_counts = self._get_allele_norm(allele_count_df = allele_count_df, thres = norm_thres)
        alleles_norm.loc[:,count_columns] = alleles_norm.loc[:,count_columns]/norm_counts
        return(alleles_norm)

    def filter_allele_counts_by_pos(self, 
        rel_pos_start = 0, rel_pos_end = 32,
        allele_uns_key = "allele_counts", filter_rel_pos = True):
        '''
        Filter alleles based on barcode matched counts, allele counts, 
        or proportion
        '''
        allele_count_df = self.uns[allele_uns_key].copy()
        filtered_allele, filtered_edits = \
            zip(*allele_count_df.allele.map(lambda a:
                filter_allele_by_pos(a, rel_pos_start, rel_pos_end, filter_rel_pos)))
        allele_count_df.loc[:, "allele"] = filtered_allele
        # Hashing on Allele object messes up the order. Converting it to str and back to allele for groupby.
        allele_count_df["str_allele"] = allele_count_df.allele.map(str)
        allele_count_df = allele_count_df.groupby(["guide", "str_allele"]).sum().reset_index()
        allele_count_df.insert(1, "allele", allele_count_df.str_allele.map(lambda s: Allele.from_str(s)))
        allele_count_df.drop("str_allele", axis=1)
        print("{} edits filtered from {} alleles.".format(sum(filtered_edits), len(filtered_edits)))
        return(allele_count_df)
        

    def collapse_allele_by_target(self, ref_base, alt_base, target_pos_column = "target_pos"):
        if not target_pos_column in self.guides.columns:
            raise ValueError("The .guides have to have 'target_pos' specifying the relative position of target edit.")
        df = self.uns["allele_counts"].copy().reset_index(drop=True)
        df["target_pos"] = self.guides.set_index("name").loc[df.guide, target_pos_column].reset_index(drop=True)
        df["has_target"] = df.apply(lambda row: row.allele.has_edit(ref_base, alt_base, rel_pos = row.target_pos), axis = 1)
        df["has_nontarget"] = df.apply(lambda row: row.allele.has_other_edit(ref_base, alt_base, rel_pos = row.target_pos), axis = 1)
        df = df.drop("target_pos", axis=1)
        res = df.groupby(["guide", "has_target", "has_nontarget"]).sum()
        return(res)

    def collapse_allele_by_nedit(self, ref_base, alt_base):
        df = self.uns["allele_counts"].copy().reset_index(drop=True)
        df["n_edits"] = df.allele.map(lambda a: sum(
            [(e.ref_base == ref_base and e.alt_base == alt_base and e.rel_pos >= 6 and e.rel_pos < 6 + 20)
            for e in a.edits]))
        df.loc[df.n_edits > 4,"n_edits"] = 4
        res = df.groupby(["guide", "n_edits"]).sum()
        return(res)



    def translate_alleles(self):
        if self.uns["allele_counts"] is None:
            print("No allele information. Run crisrpep-count with -a option.")
            return
        self.uns["allele_counts"]["aa_allele"] = self.uns["allele_counts"].allele.map(
            lambda s: get_aa_alleles(s)
        )

    def log_norm(self):
        super().log_norm()
        super().log_norm(output_layer = "lognorm_edits", read_count_layer="edits")

    def log_fold_changes(self, cond1, cond2, return_result=False):
        guide_fc = super().log_fold_change(cond1, cond2, return_result=return_result)
        edit_fc = super().log_fold_change(
            cond1,
            cond2,
            lognorm_counts_key="lognorm_edits",
            out_guides_suffix="edit_lfc",
            return_result=return_result,
        )
        if return_result:
            return (guide_fc, edit_fc)

    def log_fold_change_aggregates(
        self,
        cond1,
        cond2,
        aggregate_condit="replicate",
        compare_condit="sort",
        aggregate_fn="median",
        return_result=False,
        keep_per_replicate=False,
    ):
        guide_fc_agg = super().log_fold_change_aggregate(
            cond1,
            cond2,
            lognorm_counts_key="lognorm_counts",
            aggregate_condit=aggregate_condit,
            compare_condit=compare_condit,
            out_guides_suffix="lfc",
            aggregate_fn=aggregate_fn,
            return_result=return_result,
            keep_per_replicate=keep_per_replicate,
        )
        edit_fc_agg = super().log_fold_change_aggregate(
            cond1,
            cond2,
            lognorm_counts_key="lognorm_edits",
            aggregate_condit=aggregate_condit,
            compare_condit=compare_condit,
            out_guides_suffix="edit_lfc",
            aggregate_fn=aggregate_fn,
            return_result=return_result,
            keep_per_replicate=keep_per_replicate,
        )
        if return_result:
            return (guide_fc_agg, edit_fc_agg)

    def allele_log_fold_change_aggregate(
        self,
        cond1,
        cond2,
        aggregate_condit="replicate",
        compare_condit="sort",
        aggregate_fn="median",
        return_result=False,
        keep_per_replicate=False,
    ):
        pass

    def write(self, out_path):
        """
        Write .h5ad
        """
        adata = self.copy()
        for k in adata.uns.keys():
            if 'edit' in adata.uns[k].columns:
                adata.uns[k].edit = adata.uns[k].edit.map(str)
            for c in [colname for colname in adata.uns[k].columns if "allele" in colname]:
                adata.uns[k].loc[:,c] = adata.uns[k][c].map(str)
        super(ReporterScreen, adata).write(out_path)




def concat(screens: Collection[ReporterScreen], *args, axis = 1, **kwargs):
    # TODO: var/obs info not concated if doesn't overlap
    # TODO: concat 2 times: allele_counts merging doesn't work?
    if axis == 1:
        if not all(screen.guides.index.equals(screens[0].guides.index) for screen in screens):
            raise ValueError("Guide index doesn't match.")

    adata = ad.concat(screens, *args, axis = axis, **kwargs)
    adata.obs = screens[0].guides
    keys = set()

    # Merging uns
    if axis == 1:
        for screen in screens:
            keys.update(screen.uns.keys())

        for k in keys:
            if k == "guide_edit_counts":
                merge_on = ["guide", "edit"]
            elif k == "edit_counts":
                merge_on = ["guide", "edit"]
            elif k == "allele_counts":
                merge_on = ["guide", "allele"]
            elif k == "guide_reporter_allele_counts":
                merge_on = ["guide", "reporter_allele", "guide_allele"]
            else:
                print("uns '{}' ignored during concat.".format(k))
                continue
            for i, screen in enumerate(screens):
                if i == 0:
                    adata.uns[k] = screen.uns[k]
                else:
                    try:
                        adata.uns[k] = pd.merge(
                            adata.uns[k], screen.uns[k], on=merge_on, how="outer"
                        )
                    except:
                        print("Error occurred concatenating '{}' for {}:".format(k, screen))
                        if not k in screen.uns.keys(): print("Warning: '{}' not in this screen. Skipping...".format(k))
                        else:
                            print(k)
                            print(screen.uns[k])
            adata.uns[k] = adata.uns[k].fillna(0)
            float_col = adata.uns[k].select_dtypes(include=["float64"])
            for col in float_col.columns.values:
                adata.uns[k][col] = adata.uns[k][col].astype("int64")

    return ReporterScreen.from_adata(adata)

def read_h5ad(filename):
    adata = ad.read_h5ad(filename)
    return ReporterScreen.from_adata(adata)




from typing import List
import numpy as np
from Bio import SeqIO
from beret.framework.Edit import Edit, Allele
from beret.annotate.AminoAcidEdit import AminoAcidEdit, AminoAcidAllele, CodingNoncodingAllele
import logging
import sys

logging.basicConfig(level=logging.INFO,
                     format='%(levelname)-5s @ %(asctime)s:\n\t %(message)s \n',
                     datefmt='%a, %d %b %Y %H:%M:%S',
                     stream=sys.stderr,
                     filemode="w"
                     )
error   = logging.critical
warn    = logging.warning
debug   = logging.debug
info    = logging.info

BASE_SET = {"A", "C", "T", "G"}
reverse_map = {"A":"T", "C":"G", "G":"C", "T":"A", "N":"N", "-":"-"}

codon_map = {'TTT':'F', 'TTC':'F', 'TTA':'L', 'TTG':'L', 'CTT':'L', 'CTC':'L', 'CTA':'L', 'CTG':'L', 'ATT':'I', 'ATC':'I',
         'ATA':'I', 'ATG':'M', 'GTT':'V', 'GTC':'V', 'GTA':'V', 'GTG':'V', 'TCT':'S', 'TCC':'S', 'TCA':'S', 'TCG':'S',
         'CCT':'P', 'CCC':'P', 'CCA':'P', 'CCG':'P', 'ACT':'T', 'ACC':'T', 'ACA':'T', 'ACG':'T', 'GCT':'A', 'GCC':'A',
         'GCA':'A', 'GCG':'A', 'TAT':'Y', 'TAC':'Y', 'TAA':'*', 'TAG':'*', 'CAT':'H', 'CAC':'H', 'CAA':'Q', 'CAG':'Q',
         'AAT':'N', 'AAC':'N', 'AAA':'K', 'AAG':'K', 'GAT':'D', 'GAC':'D', 'GAA':'E', 'GAG':'E', 'TGT':'C', 'TGC':'C',
         'TGA':'*', 'TGG':'W', 'CGT':'R', 'CGC':'R', 'CGA':'R', 'CGG':'R', 'AGT':'S', 'AGC':'S', 'AGA':'R', 'AGG':'R',
         'GGT':'G', 'GGC':'G', 'GGA':'G', 'GGG':'G'}

class RefBaseMismatchException(Exception):
    pass

# This function is adopted from https://github.com/gpp-rnd/be-validation-pipeline/blob/main/notebooks/01_BEV_allele_frequencies.ipynb
def _translate(seq, codon_map):
#     if not seq: # if remove_introns returned False -> possible splice site mutation
    if seq == 'UTR':
        return 'Possible UTR mutation'
    if seq == 'intron':
        return 'Possible intronic mutation'
    aa = ''
    i = 0
    frame = 1
    while i < len(seq):
        substring = ''
        while frame <= 3:
            if i<len(seq):
                if seq[i] == '-': #print('deletion')
                    i += 1
                else:
                    substring += seq[i]  
                    i += 1
                    frame+=1
            else: # if reached end of the sequence and frame still <=3, complete codon sequence based on last_codon
                substring += last_codon[frame-1]
                i += 1
                frame+=1
        if len(substring) == 3:
            frame = 1 # reset frame 
            if ('N' in substring):
                aa = aa + '-'
            else:
                aa = aa + codon_map[substring] # translate codon
        else:
            frame = 1
    return aa

def _parse_range(chr_range):
    chrom, gen_range = chr_range.split(":")
    start, end = gen_range.split('-')
    start = int(start)
    end = int(end)
    return(chrom, start, end)
    
def _parse_description(desc_str):
    sstr = desc_str.split(" ")
    for s in sstr:
        if s.startswith("range="):
            return(_parse_range(s[6:]))

def _get_seq_pos_from_fasta(fasta_file_name: str):
    exons = list(SeqIO.parse(fasta_file_name, "fasta"))
    translated_seq = []
    genomic_pos = []
    for exon in exons:
        _, exon_start, exon_end = _parse_description(exon.description)
        for i, nt in enumerate(str(exon.seq)):
            if nt.islower(): continue
            translated_seq.append(nt)
            genomic_pos.append(exon_start + i)
    return(translated_seq, genomic_pos)

def _translate_single_codon(nt_seq_string: str, aa_pos: int) -> str:
    codon = ''.join(nt_seq_string[aa_pos*3:(aa_pos*3+3)])
    if len(codon) != 3:
        print("reached the end of CDS, frameshift.")
        return("/")
    try:
        aa = codon_map[codon]
        return(aa)
    except KeyError:
        if codon[-1] == "N" and codon[0] in BASE_SET and codon[1] in BASE_SET:
            aa_set = set()
            for N in BASE_SET:
                aa_set.add(codon_map[codon[:2] + N])
            if len(aa_set) == 1:
                return(next(iter(aa_set)))
            else:
                print("warning: no matching aa with codon {}".format(codon))
                return("_")
        else:
            print("Cannot translate codon due to ambiguity: {}".format(codon))
            return("_")
    

class CDS():  
    def __init__(self, fasta_file_name="ldlr_exons.fa"):
        self.edited_nt = type(self).nt.copy()
        self.edited_aa_pos = set()
        self.edits_noncoding = set()
        self.set_exon_fasta_name(fasta_file_name)
        
    @classmethod
    def set_exon_fasta_name(cls, fasta_file_name: str):
        cls.fasta_file_name = fasta_file_name
        cls.translated_seq, cls.genomic_pos = _get_seq_pos_from_fasta(fasta_file_name)
        cls.nt = cls.translated_seq
        cls.pos = np.array(cls.genomic_pos)
    
    def translate(self):
        self.aa = _translate(self.edited_nt, codon_map)
    
    def _get_relative_nt_pos(self, absolute_pos):
        nt_relative_pos = np.where(type(self).pos == absolute_pos)[0]
        assert len(nt_relative_pos) <= 1, nt_relative_pos
        if not nt_relative_pos: return(-1)
        return(nt_relative_pos.astype(int).item())
    
    def _edit_pos_to_aa_pos(self, edit_pos):
        nt_relative_pos = self._get_relative_nt_pos(edit_pos)
        if nt_relative_pos != -1 : 
            self.edited_aa_pos.add(nt_relative_pos // 3)
        return(nt_relative_pos)
    
    def edit_single(self, edit_str):
        edit = Edit.from_str(edit_str)
        rel_pos= self._edit_pos_to_aa_pos(edit.pos)
        if edit.strand == '-': 
            ref_base = reverse_map[edit.ref_base]
            alt_base = reverse_map[edit.alt_base]
        else:
            ref_base = edit.ref_base
            alt_base = edit.alt_base
        if rel_pos == -1 : # position outside CDS
            self.edits_noncoding.add(edit)
        elif type(self).nt[rel_pos] != ref_base:
            raise RefBaseMismatchException("ref:{} at pos {}, got edit {}".format(type(self).nt[rel_pos], rel_pos, edit))
        else:
            self.edited_nt[rel_pos] = alt_base
        if alt_base == '-': #frameshift
            #self.edited_nt.pop(rel_pos)
            self.edited_aa_pos.update(list(range(rel_pos, len(CDS.nt) // 3)))
        
    def edit_allele(self, allele_str):
        if isinstance(allele_str, Allele):
            edit_strs = allele_str.edits
        else: edit_strs = allele_str.split(",")
        for edit_str in edit_strs:
            self.edit_single(edit_str)
        if '-' in self.edited_nt:
            self.edited_nt.remove('-')
    
    def get_aa_change(self, include_synonymous = True) -> List[str]:
        mutations = CodingNoncodingAllele()
        mutations.nt_allele.update(self.edits_noncoding)
        for edited_aa_pos in self.edited_aa_pos:
            ref_aa = _translate_single_codon(type(self).nt, edited_aa_pos)
            mt_aa = _translate_single_codon(self.edited_nt, edited_aa_pos)
            if mt_aa == "_": 
                return("translation error")
            if not include_synonymous and ref_aa == mt_aa:
                continue
            mutations.aa_allele.add(AminoAcidEdit(
                edited_aa_pos + 1, ref_aa, mt_aa))
        
        return(mutations)
   



from typing import List, Tuple
from .Edit import Edit, Allele
from ..annotate.translate_allele import CDS, RefBaseMismatchException


def filter_allele_by_pos(
    allele: Allele,
    pos_start: int = None, 
    pos_end: int = None,
    filter_rel_pos = True
    ):
    '''
    Filter alleles based on position and return the filtered allele and
    number of filtered edits.
    '''
    filtered_edits = 0
    
    if not (pos_start is None and pos_end is None):
        if filter_rel_pos:
            for edit in allele.edits.copy():
                if not (edit.rel_pos >= pos_start and edit.rel_pos < pos_end): 
                    filtered_edits += 1
                    allele.edits.remove(edit)
        else:
            for edit in allele.edits.copy():
                if not (edit.pos >= pos_start and edit.pos < pos_end): 
                    filtered_edits += 1
                    allele.edits.remove(edit)
    else:
        print("No threshold specified") # TODO: warn
    return(allele, filtered_edits)

def filter_allele_by_base(
    allele: Allele,
    allowed_base_changes: List[Tuple] = None,
    allowed_ref_base: str = None, allowed_alt_base: str = None
):
    '''
    Filter alleles based on position and return the filtered allele and
    number of filtered edits.
    '''
    filtered_edits = 0
    if not (allowed_ref_base is None and allowed_alt_base is None) + (allowed_base_changes is None) == 1:
        print("No filters specified or misspecified filters.")
    elif not allowed_base_changes is None:
        for edit in allele.edits.copy():
            if not (edit.ref_base, edit.alt_base) in allowed_base_changes: 
                filtered_edits += 1
                allele.edits.remove(edit)
    elif not allowed_ref_base is None:
        for edit in allele.edits.copy():
            if edit.ref_base != allowed_ref_base: 
                filtered_edits += 1
                allele.edits.remove(edit)
            elif not allowed_alt_base is None and edit.alt_base != allowed_alt_base:
                filtered_edits += 1
                allele.edits.remove(edit)
    else:
        for edit in allele.edits.copy():
            if edit.alt_base != allowed_alt_base: 
                filtered_edits += 1
                allele.edits.remove(edit)
    return(allele, filtered_edits)

def get_aa_alleles(allele_str, include_synonymous = True):
    ldlr_cds = CDS()
    try:
        ldlr_cds.edit_allele(allele_str)
        ldlr_cds.get_aa_change(include_synonymous)
    except RefBaseMismatchException as e:
        print(e)
        return("ref mismatch")
    return ldlr_cds.get_aa_change(True)


from enum import unique
from typing import Iterable

class Edit:
    reverse_map = {"A":"T", "C":"G", "T":"A", "G":"C", "-":"-"}
    strand_map = {"+":1, "-":-1}
    def __init__(self, rel_pos: int, ref_base: chr, alt_base: chr, offset: int = None, strand: int = 1, 
    unique_identifier = None):
        assert strand in [+1, -1]
        strand_to_symbol = {1:'+', -1:'-'}
        self.rel_pos = rel_pos  
        self.ref_base = ref_base  # TODO make it ref / alt instead of ref_base and alt_base for AAEdit comp. or make abstract class
        self.alt_base = alt_base
        self.uid = unique_identifier
        if type(strand) == int:
            self.strand = strand_to_symbol[strand]
        else:
            assert strand in ["+", "-"]
            self.strand = strand
        if not offset is None:
            self.pos = offset + self.rel_pos * strand
        else:
            self.pos = self.rel_pos

    @classmethod
    def from_str(cls, edit_str): #pos:strand:start>end
        if type(edit_str) is Edit: return(edit_str)
        pos, rel_pos, strand, base_change = edit_str.split(":")
        pos = int(pos)
        rel_pos = int(rel_pos)
        assert strand in ["+", "-"]
        strand = cls.strand_map[strand]
        offset = pos - rel_pos*strand
        ref_base, alt_base = base_change.split(">")
        
        # if strand == "-":
        #     start = cls.reverse_map[start]
        #     end = cls.reverse_map[end]
        return(cls(rel_pos, ref_base, alt_base, offset = offset, strand = strand))

    def get_abs_edit(self):
        '''
        Returns absolute edit representation regardless of the relative edit position by guide.
        '''
        if self.strand == '-':
            ref_base = type(self).reverse_map[self.ref_base]
            alt_base = type(self).reverse_map[self.alt_base]
        else:
            ref_base = self.ref_base
            alt_base = self.alt_base
        if not self.uid is None:
            return("{}!{}:{}>{}".format(self.uid, int(self.pos), ref_base, alt_base))
        return("{}:{}>{}".format(int(self.pos), ref_base, alt_base))

    def set_uid(self, uid):
        self.uid = uid
        return(self)

    def __eq__(self, other):
        if (
            self.pos == other.pos
            and self.ref_base == other.ref_base
            and self.alt_base == other.alt_base
        ):
            return True
        return False
    
    def __lt__(self, other): # Implemented for pandas compatibility
        return self.pos < other.pos

    def __gt__(self, other): # Implemented for pandas compatibility
        if isinstance(other, Edit):
            return self.pos > other.pos
        else:
            return self.__repr__() > str(other)

    def __hash__(self):
        # Note that this doesn't include relative bases. 
        # This wouldn't matter if we assign edit to each guide.
        return(hash((self.pos, self.ref_base, self.alt_base)))

    def __repr__(self):
        return(("{}:{}:{}:{}>{}".format(
            int(self.pos), 
            int(self.rel_pos),
            self.strand, 
            self.ref_base, 
            self.alt_base)))


class Allele:
    # pos, ref, alt
    def __init__(self, edits: Iterable[Edit] = None):
        if edits is None:
            self.edits = set()
        else:
            self.edits = set(edits)

    @classmethod
    def from_str(cls, allele_str): #pos:strand:start>end
        if type(allele_str) is Allele: return(allele_str)
        edits = set()
        try:
            for edit_str in allele_str.split(","):
                edit = Edit.from_str(edit_str)
                edits.add(edit)
        except ValueError:
            if allele_str.strip() == '':
                return(cls(None))
        return(cls(edits))

    def has_edit(self, ref_base, alt_base, pos = None, rel_pos = None):
        if not (pos is None) + (rel_pos is None): 
            raise ValueError("Either pos or rel_pos should be specified")
        
        for e in self.edits:
            if e.ref_base == ref_base and e.alt_base == alt_base:
                if not pos is None:
                    if e.pos == pos: return True
                elif e.rel_pos == rel_pos: return True
        return False

    def has_other_edit(self, ref_base, alt_base, pos = None, rel_pos = None):
        '''
        Has other edit than specified.
        '''
        if len(self.edits) == 0: return False
        if not (pos is None) + (rel_pos is None): 
            raise ValueError("Either pos or rel_pos should be specified")
        for e in self.edits:
            if e.ref_base == ref_base and e.alt_base == alt_base:
                if not pos is None:
                    if e.pos != pos: return True
                elif e.rel_pos != rel_pos: return True
            else: return True
        return False

    def __eq__(self, other):
        if self.__repr__() == other.__repr__():
            return True
        return False

    def __lt__(self, other): # Implemented for pandas compatibility
        return len(self.edits) < len(other.edits)

    def __hash__(self):
        return(hash(self.__repr__()))

    def add(self, edit: Edit):
        self.edits.add(edit)  # TBD: adding to set?

    def update(self, edits: Iterable[Edit]):
        self.edits.update(edits)

    def __repr__(self):
        if len(self.edits) == 0: return ""
        list_edits = sorted(list(self.edits.copy()))
        list_edits = list(map(lambda s: str(s), list_edits))
        return(",".join(list_edits))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
     name='berets',  
     version='0.0.8',
     author="Jayoung Ryu",
     author_email="jayoung_ryu@g.harvard.edu",
     description="Tools for analyzing CRISPR data with REPorter edits",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://github.com/pinellolab/beret",
     packages=setuptools.find_packages(),
     scripts=["bin/beret-count",
     "bin/beret-count-samples"],
     install_requires=[
        'numpy',
        'perturb-tools>=0.0.16',
      ],
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
 )




# -*- coding: utf-8 -*-

import os
import gzip
import argparse
import sys
import gzip
import subprocess as sb
from collections import defaultdict
import unicodedata
import re
import pandas as pd
import numpy as np
from Bio import SeqIO
from Bio.SeqIO.QualityIO import FastqGeneralIterator
from Bio.Seq import Seq
from typing import Dict, List, Union

import logging
logging.basicConfig(level=logging.INFO,
                     format='%(levelname)-5s @ %(asctime)s:\n\t %(message)s \n',
                     datefmt='%a, %d %b %Y %H:%M:%S',
                     stream=sys.stderr,
                     filemode="w"
                     )
error   = logging.critical
warn    = logging.warning
debug   = logging.debug
info    = logging.info


_ROOT = os.path.abspath(os.path.dirname(__file__))


####Support functions###

def check_file(filename):
    try:
        with open(filename): pass
    except IOError:
        raise Exception('I cannot open the file: '+filename)
 

def check_library(library_name):
        try:
                return __import__(library_name)
        except:
                error('You need to install %s module to use CRISPRessoCount!' % library_name)
                sys.exit(1)


def slugify(value): #adapted from the Django project
    
    value = unicodedata.normalize('NFKD', unicode(value)).encode('ascii', 'ignore')
    value = unicode(re.sub('[^\w\s-]', '_', value).strip())
    value = unicode(re.sub('[-\s]+', '-', value))
    
    return str(value)


def get_fastq_handle(fastq_filename):
    if fastq_filename.endswith('.gz'):
            fastq_handle=gzip.open(fastq_filename)
    else:
            fastq_handle=open(fastq_filename)
    return(fastq_handle)


def read_is_good_quality(record: SeqIO.SeqRecord, 
        min_bp_quality = 0, 
        min_single_bp_quality = 0, 
        qend = -1):
    mean_quality_pass = np.array(record.letter_annotations["phred_quality"])[:qend_R1].mean() >= min_bp_quality
    min_quality_pass = np.array(record.letter_annotations["phred_quality"])[:qend_R1].min()>=min_single_bp_quality
    return(mean_quality_pass and min_quality_pass)


def check_read_names_and_filter_quality(R1_filename, 
        R2_filename, 
        output_R1_filename=None, 
        output_R2_filename = None,
        filter_by_qual = False,
        min_bp_quality=20,
        min_single_bp_quality=0, 
        qend_R1 = -1, 
        qend_R2 = -1):
        
        if min_bp_quality > 0 or min_single_bp_quality > 0:
            info('Filtering reads with average bp quality < {} \
                and single bp quality < {} ... in up to position {} of R1 and {} of R2'.format(args.min_average_read_quality,
                    args.min_single_bp_quality))

        if qend_R1 > 0 or qend_R2 > 0:
            info("In the filtering, bases up to position {} of R1 and {} of R2 are considered.".format(qend_R1, qend_R2))

        R1_handle = get_fastq_handle(R1_filename)
        R2_handle = get_fastq_handle(R2_filename)

        if not output_R1_filename:
                output_R1_filename = R1_filename.replace('.fastq', '').replace('.gz', '') + '_filtered.fastq.gz'
                output_R2_filename = R2_filename.replace('.fastq', '').replace('.gz', '') + '_filtered.fastq.gz'

        try:
            if filter_by_qual:
                R1_filtered = gzip.open(output_R1_filename, 'w+')
                R2_filtered = gzip.open(output_R2_filename, 'w+')
            R1 = list(SeqIO.parse(R1_handle, "fastq"))
            R2 = list(SeqIO.parse(R2_handle, "fastq"))
            R1_handle.close()
            R2_handle.close()

            if len(R1) != len(R2):
                raise ValueError("The number of reads in R1 and R2 file does not match.")

            for i in range(len(R1)):
                R1_record = R1[i]
                R2_record = R2[i]

                if R1_record.name != R2_record.name : 
                    raise InputFileError("R1 and R2 read discordance in read {} and {}".format(R1_record.name, R2_record.name))
                if filter_by_qual:
                    R1_quality_pass = read_is_good_quality(R1_record, min_bp_quality, min_single_bp_quality, qend_R1)
                    R2_quality_pass = read_is_good_quality(R2_record, min_bp_quality, min_single_bp_quality, qend_R2)

                    if R1_quality_pass and R2_quality_pass:
                        R1_filtered.write(record.format('fastq'))
                        R2_filtered.write(record.format('fastq'))
        except:
                raise Exception('Error handling the fastq_filtered_outfile')

        if filter_by_qual: 
            return((output_R1_filename, output_R2_filename))
        else:
            return((R1_filename, R2_filename))


def find_wrong_nt(sequence):
    return(list(set(sequence.upper()).difference(set(['A','T','C','G','N']))))


def get_n_reads_fastq(fastq_filename):
     p = sb.Popen(('z' if fastq_filename.endswith('.gz') else '' ) +"cat < %s | wc -l" % fastq_filename , shell=True,stdout=sb.PIPE)
     return(int(float(p.communicate()[0])/4.0))


def mask_sequence_positions(seq: str, pos: np.array) -> str:
    return("".join([seq[i] if i in pos else "N" for i in range(len(seq))]))


def revcomp(seq: str) -> str:
    seq = Seq(seq)
    return(str(seq.reverse_complement()))


###EXCEPTIONS############################
class NTException(Exception):
    pass

class InputFileError(Exception):
    pass

class NoReadsAfterQualityFiltering(Exception):
    pass
#########################################

def get_input_parser():
    """Get the input data"""
    print('  \n~~~CRISPRessoCount~~~')
    print('-Utility to perform sgRNA enumeration from deep sequencing data-')
    print(r'''
          )                                             )
         (           ________________________          (
        __)__       | __   __            ___ |        __)__
     C\|     \      |/  ` /  \ |  | |\ |  |  |     C\|     \
       \     /      |\__, \__/ \__/ | \|  |  |       \     /
        \___/       |________________________|        \___/
    ''')
    
    
    print('\n[Luca Pinello 2017, send bugs, suggestions or *green coffee* to lucapinello AT gmail DOT com]\n\n')
    
    
    parser = argparse.ArgumentParser(description='CRISPRessoCount parameters',formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--R1', type=str,  help='fastq file for read 1', required=True,default='Fastq filename' )
    parser.add_argument('--R2', type=str,  help='fastq file for read 2, sorted as the same name order as in --R1 file.', required=True, default='Fastq filename' )
    parser.add_argument('-b', '--edited_base', type = str, required = True, help = 'For base editors, the base that should be ignored when matching the gRNA sequence')
    parser.add_argument('-f','--sgRNA_file', type=str, required = True, help='''sgRNA description file. The format requires three columns: gRNA, Reporter, gRNA_barcode.''')

    #optional
    parser.add_argument('--guide_start', type = int, help = "Starting position of guide in R1", default = 24)
    parser.add_argument('-r', '--count_reporter', type = bool, help = "Count reporter edits.", default = True)
    parser.add_argument('-q','--min_average_read_quality', type=int, help='Minimum average quality score (phred33) to keep a read', default=0)
    parser.add_argument('-s','--min_single_bp_quality', type=int, help='Minimum single bp score (phred33) to keep a read', default=0)
    parser.add_argument('-n','--name',  help='Output name', default='')
    parser.add_argument('-o','--output_folder',  help='', default='')
    parser.add_argument('-l', '--reporter_length', type = int, help = "length of the reporter", default = 32)
    parser.add_argument('--keep_intermediate',help='Keep all the  intermediate files',action='store_true')
    parser.add_argument('--qstart_R1', help='Start position of the read when filtering for quality score of the read 1', type = int, default = 0)
    parser.add_argument('--qend_R1', help = 'End position of the read when filtering for quality score of the read 1', type = int, default = 47)
    parser.add_argument('--qstart_R2', help = 'Same as qstart_R1, for read 2 fastq file', default = 0)
    parser.add_argument('--qend_R2', help = 'Same as qstart_R2, for read 2 fastq file', default = 36)
    parser.add_argument('--guide_bc_len', help = 'Guide barcode sequence length at the beginning of the R2', type = str, default = 4)

    return(parser)


def check_arguments(args):
    """Check the argument validity of the ArgumentParser"""
    check_file(args.R1)
    check_file(args.R2)

    if args.sgRNA_file:
        check_file(args.sgRNA_file)

    # Edited base should be one of A/C/T/G
    if args.edited_base.upper() not in ['A', 'C', "T", "G"]:
        raise ValueError("The edited base should be one of A/C/T/G, {} provided.".format(args.edited_base))
    else:
        edited_base = args.edited_base.upper()
    info('Using specified edited base: %s' % edited_base)

    read_length = get_first_read_length(args.R1)

    # Check if positions of guide and quality control is valid
    if not (args.guide_start >= 0 and args.guide_start + max(GUIDE_LENGTHS) < read_length): 
        raise ValueError("The guide start and end position is not valid. Start position should be nonnegative integer ({} provided), and the ending position should be shorter than the read length {} ({} provided)".format(args.guide_start, read_length, args.guide_start + max(GUIDE_LENGTHS)))
    
    if not (args.qstart_R1 < read_length and args.qstart_R2 < read_length):
        raise ValueError("The start position of base quality filter is not nonnegative ({} for R1, {} for R2 provided)".format(args.qstart_R1, args.qstart_R2))
    
    if not (args.qend_R1 < read_length and args.qend_R2 < read_length):
        raise ValueError("The start position of base quality filter is not nonnegative ({} for R1, {} for R2 provided)".format(args.qstart_R1, args.qstart_R2))

    if args.qend_R2 != args.guide_bc_len + args.reporter_length :
        warn("Quality of R2 checked up until {}bp, while the length of guide barcode and reporter combined is {}bp.".format(
            args.qend_R2, args.guide_bc_len + args.reporter_length))
    
    #normalize name and remove not allowed characters
    if args.name:   
        clean_name=slugify(args.name)
        if args.name!= clean_name:
               warn('The specified name %s contained characters not allowed and was changed to: %s' % (args.name,clean_name))
               args.name=clean_name
    info("Done checking input arguments.")

    return(args)

    #count reads
#def check_input_reads(args):
#    N_READS_INPUT=get_n_reads_fastq(args.R1)
#    N_READS_INPUT_R2 = get_n_reads_fastq(args.R2)
#    if N_READS_INPUT != N_READS_INPUT_R2 : 
#        raise InputFileError("Number of reads not matched in two input fastq file")
#    return(N_READS_INPUT)



def get_database_id(args):    
    get_name_from_fasta = lambda  x: os.path.basename(x).replace('.fastq','').replace('.gz','').replace("_R1",'')
    
    if not args.name:
        database_id='%s' % get_name_from_fasta(args.R1)    
    else:
        database_id=args.name
    return(database_id)


def read_count_match(R1_filename, R2_filename) -> int:
    R1_count = get_n_reads_fastq(R1_filename)
    R2_count = get_n_reads_fastq(R2_filename)
    if R1_count != R2_count: 
        raise InputFileError("Paired end read numbers are different in R1({}) and R2({})".format(R1_count, R2_count)) 
    return(R1_count)


def get_first_read_length(fastq_filename):
    for record in SeqIO.parse(fastq_filename, "fastq"):
        return(len(record))
    raise InputFileError("Provided R1 file doesn't have any read to parse")


def get_sgRNA_dict(sgRNA_filename: str, edited_base: str) -> Dict[str, Dict[np.array, Dict[str, str]]]:
    """Reads the sgRNA file (csv) with three columns: name (of gRNA), gRNA(sequence), gRNA_barcode.
    Will create a dictionary of (barcode -> (masked position -> (masked sequence -> gRNA name))) 
    for fast search of gRNA match with masked positon.
    """

    info('Using guides information in %s' % sgRNA_filename)
    masked_guides_dict = dict()
    guides = dict()

    with open(sgRNA_filename) as infile:
        sgRNA_df = pd.read_csv(infile)
        if not ('gRNA' in sgRNA_df.columns and 'gRNA_barcode' in sgRNA_df.columns):
            raise InputFileError("Input gRNA file doesn't have the column 'gRNA' or 'gRNA_barcode'.")
        
        duplicate_gRNA = 0
        for i, guide in enumerate(sgRNA_df["gRNA"]):
            if not len(guide) in GUIDE_LENGTHS:
                warn("Guide length does not match the prespecified length ({}bp)".format(",".join(GUIDE_LENGTHS)))
            guide = guide.replace(edited_base, "N")
            
            # Writing gRNA stats with duplication, masked gRNA
            if guide in guides.keys(): 
                guides[guide] += 1
            else: guides[guide] = 1

            gRNA_bc = sgRNA_df["gRNA_barcode"][i]
            gRNA_seq = sgRNA_df["gRNA"][i]
            mask_pos = tuple(np.nonzero(gRNA_seq == args.edited_base)[0])  # Indices of the masked position (np.array)
            masked_seq = gRNA_seq.replace(edited_base, "N")
            gRNA_name = sgRNA_df["name"][i]

            if gRNA_bc in masked_guides_dict.keys():
                if mask_pos in masked_guides_dict[gRNA_bc].keys():
                    if masked_seq in masked_guides_dict[gRNA_bc][mask_pos].keys():
                        duplicate_gRNA += 1
                    else:
                        masked_guides_dict[gRNA_bc][mask_pos][masked_seq] = gRNA_name
                else:
                    masked_guides_dict[gRNA_bc][mask_pos] = {masked_seq: gRNA_name}                    
            else: 
                masked_guides_dict[gRNA_bc] = {mask_pos : {masked_seq: gRNA_name}}
    return(masked_guides_dict)


def get_guide_to_reporter_dict(sgRNA_filename: str) -> Dict[str, str]:
    '''Returns a gRNA name to reporter sequence mapping.'''
    guide_to_reporter = {}
    with open(sgRNA_filename) as infile:
        sgRNA_df = pd.read_csv(infile)
        if not ('name' in sgRNA_df.columns and 'Reporter' in sgRNA_df.columns):
            raise InputFileError("Input gRNA file doesn't have the column 'gRNA' or 'gRNA_barcode'.")
        return(sgRNA_df[["name", "Reporter"]].set_index('name').to_dict())


def match_masked_sgRNA(masked_guides_dict: dict, 
    gRNA_seq: str, 
    gRNA_barcode: str) -> List[str]:
    ''' 
    Match a gRNA_seq to the masked_guides_dict

    masked_guides_dict : dict
        (barcode -> (masked position -> (masked sequence -> gRNA name)))
    '''
    n_matches = 0
    matches = []
    if gRNA_barcode in masked_guides_dict.keys():
        for mask, maseq2name in masked_guides_dict[gRNA_barcode].items():   
            masked_seq = "".join(["N" if i in mask else gRNA_seq[i] for i in range(len(gRNA_seq))])
            # maseq2name : (masked sequence -> gRNA name)
            if masked_seq in maseq2name.keys():
                n_matches += 1
                matches.append(maseq2name[masked_seq])
            else:
                pass # no match with current mask. Try other mask
    else:
        pass # no barcode match.
    return(matches)


def count_masked_guides(R1_filename, R2_filename, 
    masked_guides_dict: Dict[str, Dict[np.array, Dict[str, str]]], 
    guide_start: int, 
    guide_bc_len: int, 
    write_nomatch: bool = False, 
    count_reporter_edits: bool = False, 
    guide_to_reporter_dict: dict = None) -> Union[Dict[str, int], Dict[str, Dict[tuple, int]]]:
    '''
    Given a read pair, find matches among gRNA and optionally find the reporter edits.
    Returns a dictionary of (guide name -> count) if count_reporter_edits = False.
    Returns a dictionary of (guide name -> ((pos, ref_base, edited_base) -> count)) otherwise.
    '''

    infile_R1 = get_fastq_handle(R1_filename)
    infile_R2 = get_fastq_handle(R2_filename)

    if write_nomatch:
        nomatch_R1_filename = R1_filename.replace(".fastq", "_nomatch.fastq")
        nomatch_R2_filename = R2_filename.replace(".fastq", "_nomatch.fastq")
        outfile_R1_nomatch = open(nomatch_R1_filename, "w")
        outfile_R2_nomatch = open(nomatch_R2_filename, "w")

    if count_reporter_edits:
        assert not guide_to_reporter_dict is None, "No guide to reporter dictionary passed onto count_masked_guides."

    info('Counting sgRNAs...')
    
    iterator_R1 = FastqGeneralIterator(infile_R1)
    iterator_R2 = FastqGeneralIterator(infile_R2)

    gname_to_count = dict()

    for r1, r2 in zip(iterator_R1, iterator_R2):
        t1, R1_seq, q1 = r1
        t2, R2_seq, q2 = r2
        gRNA_names = []
        for guide_length in GUIDE_LENGTHS:
            gRNA_barcode = revcomp(R2_seq[:guide_bc_len])
            gRNA_seq = R1_seq[guide_start:guide_start + guide_length]
            gRNA_names.extend(match_masked_sgRNA(masked_guides_dict, gRNA_seq, gRNA_barcode))

        if len(gRNA_names) == 0: 
            if write_nomatch:
                outfile_R1_nomatch.write("{}\n{}\n+\n{}\n".format(t1, R1_seq, q1))
                outfile_R2_nomatch.write("{}\n{}\n+\n{}\n".format(t2, R2_seq, q2))
        elif len(gRNA_names) > 1: warn("{}th read with {} nonunique mapping. Discarding the read.".format(i, len(gRNA_names)))
        else:
            # unique match
            gRNA_name = gRNA_names[0]

            if count_reporter_edits:
                if gRNA_name in gname_to_count.keys():
                    reporter_edit_counts = gname_to_count[gRNA_name]
                else: 
                    reporter_edit_counts = dict()
                reporter_ref = guide_to_reporter_dict[gRNA_name]
                reporter_seq = revcomp(R2_seq[guide_bc_len:guide_bc_len + REPORTER_LENGTH])
                for i, (ref_nt, sample_nt) in enumerate(zip(reporter_ref, reporter_seq)):
                    if ref_nt == sample_nt: continue
                    else: 
                        edit = (i, ref_nt, sample_nt)
                        if edit in reporter_edit_counts.keys():
                            reporter_edit_counts[edit] += 1
                        else: 
                            reporter_edit_counts[edit] = 1
                gname_to_count[gRNA_name] = reporter_edit_counts

            else:
                # Only count the gRNA counts
                if gRNA_name in gname_to_count.keys(): 
                    gname_to_count[gRNA_name] += 1
                else: 
                    gname_to_count[gRNA_name] = 1

    return(gname_to_count)


if __name__ == '__main__':
        REPORTER_LENGTH = 32
        GUIDE_LENGTHS = [19, 20]
#    try:
        parser = get_input_parser()
        args = parser.parse_args()

        args = check_arguments(args)
        N_READS_INPUT = read_count_match(args.R1, args.R2)

        database_id = get_database_id(args)
        OUTPUT_DIRECTORY='beretCount_on_%s' % database_id
        if args.output_folder:
                OUTPUT_DIRECTORY=os.path.join(os.path.abspath(args.output_folder),OUTPUT_DIRECTORY)
        
        _jp=lambda filename: os.path.join(OUTPUT_DIRECTORY,filename) #handy function to put a file in the output directory
        log_filename=_jp('beretCount_RUNNING_LOG.txt')
        
        
        try:
             os.makedirs(OUTPUT_DIRECTORY)
             logging.getLogger().addHandler(logging.FileHandler(log_filename))
    
             with open(log_filename,'w+') as outfile:
                 outfile.write('[Command used]:\nCRISPRessoCount %s\n\n[Execution log]:\n' % ' '.join(sys.argv))
             info('Creating Folder %s' % OUTPUT_DIRECTORY)
             info('Done!')
        except:
                 warn('Folder %s already exists.' % OUTPUT_DIRECTORY)


        # Check read name is aligned to be the same in R1 and R2 files.
        # Optionally filter the reads based on the base quality

        filtered_R1_filename, filtered_R2_filename = check_read_names_and_filter_quality(args.R1,
            args.R2, 
            output_R1_filename=_jp(os.path.basename(args.R1).replace('.fastq','').replace('.gz','')+'_filtered.fastq.gz'),
            output_R2_filename=_jp(os.path.basename(args.R2).replace('.fastq','').replace('.gz','')+'_filtered.fastq.gz'),
            min_bp_quality=args.min_average_read_quality,
            min_single_bp_quality=args.min_single_bp_quality, 
            qend_R1 = args.qend_R1, 
            qend_R2 = args.qend_R2)

        info('Done!')
        
        N_READS_AFTER_PREPROCESSING = read_count_match(filtered_R1_filename, filtered_R2_filename)
        
        if N_READS_AFTER_PREPROCESSING == 0:             
            raise NoReadsAfterQualityFiltering('No reads in input or no reads survived the average or single bp quality filtering.')
        else:
            info('Number of reads in input:%d\tNumber of reads after filtering:%d\n' % (N_READS_INPUT, N_READS_AFTER_PREPROCESSING))
 
        masked_guides_dict = get_sgRNA_dict(args.sgRNA_file, args.edited_base)

        if args.count_reporter: 
            guide_to_reporter = get_guide_to_reporter_dict(args.sgRNA_file)
            gRNA_count = count_masked_guides(filtered_R1_filename, 
                filtered_R2_filename, 
                masked_guides_dict, 
                args.guide_start, 
                args.guide_bc_len, 
                write_nomatch = True, 
                count_reporter_edits = True, 
                guide_to_reporter_dict = guide_to_reporter)
        else:
            gRNA_count = count_masked_guides(filtered_R1_filename, 
                filtered_R2_filename, 
                masked_guides_dict, 
                args.guide_start, 
                args.guide_bc_len, 
                write_nomatch = True, 
                count_reporter_edits = False, 
                )

        info('Done!')    

        info('Writing output table...')

        if args.count_reporter:
            df_guide_counts = pd.DataFrame.from_records(
                [(name, pos, ref_base, sample_base, count)
                for name, edit_to_count in gRNA_count.items()
                for (pos, ref_base, sample_base), count in edit_to_count.items() 
                ],
                columns = ["name", "pos", "ref_base", "sample_base", "count"]
            )
            df_guide_counts.sort_values(by = "count").to_csv(
                _jp('readCount_edit_{}.txt'.format(database_id)), sep='\t')

        else:
            df_guide_counts = pd.Series(gRNA_count, name = "read_counts").to_frame()
            df_guide_counts.index.name='guide_name'
            df_guide_counts['read_%']=df_guide_counts['Read_Counts']/N_READS*100
            df_guide_counts['RPM']=df_guide_counts['Read_Counts']/N_READS*1000000
            df_guide_counts.sort_values(by='Read_Counts',ascending=False).to_csv(
                _jp('readCount_guide_{}.txt'.format(database_id), sep='\t'))
        
        info('Done!')
        
        if not args.keep_intermediate:
            info('Removing Intermediate files...')
            
            files_to_remove=[]
                             
            if args.min_average_read_quality>0 or args.min_single_bp_quality>0:
                files_to_remove.extend([filtered_R1_filename, filtered_R2_filename])
                
            for file_to_remove in files_to_remove:
                 try:
                     if os.path.islink(file_to_remove):
                         os.unlink(file_to_remove)
                     else:                             
                         os.remove(file_to_remove)
                 except:
                     warn('Skipping:%s' %file_to_remove)    

 
        info('All Done!')
        print(r'''
              )                                             )
             (           ________________________          (
            __)__       | __   __            ___ |        __)__
         C\|     \      |/  ` /  \ |  | |\ |  |  |     C\|     \
           \     /      |\__, \__/ \__/ | \|  |  |       \     /
            \___/       |________________________|        \___/
        ''')
        sys.exit(0)
    
#    except Exception as e:
#        error('\n\nERROR: %s' % e)
#        sys.exit(-1)

